"""
Dataset tools
"""
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"
__license__ = "GPL"

import pandas as pd
from pathlib import Path
import panel as pn
import param
import copy
import warnings
import os
from panel.viewable import Viewer
from io import BytesIO

from pyDendron.dataset import Dataset
from pyDendron.alien.io_besancon import IOBesancon
from pyDendron.alien.io_heidelberg import IOHeidelberg
from pyDendron.alien.io_rwl import IORWL
from pyDendron.alien.io_sylphe import IOSylphe
from pyDendron.alien.io_dendronIV import IODendronIV
from pyDendron.alien.io_tridas import IOTridas
from pyDendron.tools.alignment import Alignment
from pyDendron.dataname import *
from pyDendron.gui_panel.dataset_package import DatasetPackage
from pyDendron.gui_panel.tabulator import (_cell_text_align, _cell_editors, _header_filters,
                                           _cell_formatters, unique_filename,
                                           )

from pyDendron.app_logger import logger, notification_stream, perror
from pyDendron.gui_panel.my_viewer import MyViewer
from pyDendron.mean import data2col


class ToolsPanel(MyViewer):
    """ Tools to manage dataset. """
    path = param.Foldername('./', doc='Output path of the saved package.')
    filters = param.List(['*.p', '*.xlsx', '*.json'], doc='glob filter')
    encoding = param.Selector(default='utf-8', objects=['utf-8', 'latin1'], doc='file char encoding')
    tolerance = param.Integer(default=5, bounds=(1, 10), step=1, doc='Number of equal values tolerated.')
        
    def __init__(self, dataset_selector, dataset, parameters, dendron_info, cfg_path, tmp_path, **params):
        super().__init__(cfg_path, **params)
        self.dataset = dataset
        self.parameters = parameters
        self.tmp_path = tmp_path
        #self.wcolumns = parameters.columns.get_widgets()
        self.param_columns = parameters.column_package

        self.dataset_package = DatasetPackage(dataset, parameters, name='tools')
        self.dendron_info = dendron_info
        self.path = self.dendron_info.dataset_path
        self.www = self.dendron_info.www
        self.import_dataset = None
        self.outliers = []
        self.dataset_selector = dataset_selector
        
        self.bt_size = 90

        self._layout = pn.Tabs(
            ('Import', self._import()),
            ('Export', self._export()),
            ('Merge', self._merge()),
            ('Detect', self._validate()),
            ('Edit rings', self._edit_ring()),
            name=self.name,
             dynamic=True, styles={'font-size': '15px'})

    def get_sidebar(self, visible=True):   
        return pn.Card(pn.Param(self, show_name=False),                 
                margin=(5, 0), 
                sizing_mode='stretch_width', 
                title='Tools',
                collapsed=True, visible=visible)
  
    def get_options(self):
        """ Retrieve file options based on filters. """
        options = {}
        for flt in self.filters:
            for file in Path(self.dendron_info.dataset_path).glob(flt):
                options[f'\U0001F4E6 {str(file.name)}'] = file
        return options
    
    def _import(self):

        def on_import(event):               
            """ Handle importing of datasets. """  
            def get_input_io(fn):
                tmp = fn.lower()
                if tmp.endswith('.mdb'):
                    return IOSylphe(base_directory=self.tmp_path)
                elif tmp.endswith('.fh') or tmp.endswith('.fh.zip'):
                    return IOHeidelberg()
                elif tmp.endswith('.rwl') or tmp.endswith('.crn') or tmp.endswith('.crn.zip') or tmp.endswith('.rwl.zip'):
                    return IORWL()
                elif tmp.endswith('.bes') or tmp.endswith('.bes.zip'):
                    return IOBesancon()
                elif tmp.endswith('.dxml.zip'):
                    return IODendronIV(base_directory=self.tmp_path)
                elif tmp.endswith('.tridas.zip') or tmp.endswith('tridas.xml'):
                    return IOTridas(base_directory=self.tmp_path)
                elif tmp.endswith('.p') or tmp.endswith('.json') or tmp.endswith('.xlsx'):
                    return Dataset(cfg_tmp=self.tmp_path)
                raise ValueError(f'Unknown file format {fn}, must be: mdb, fh, fh.zip, rwl, rwl.zip, crn, crn.zip, bes, bes.zip, Dxml.zip, tridas.zip, p, json, xlsx.')
            
            self._layout.loading = True
            try:    
                for fn, buffer in file_input.value.items():
                    io_import = get_input_io(fn)
                    parent_keycode = Path(fn).stem
                    notifications = pn.config.notifications
                    pn.config.notifications = False
                    ds = io_import.read_buffer(parent_keycode, buffer, file_input.mime_type[fn])
                    if self.import_dataset is None:
                        self.import_dataset = ds  
                    else:
                        self.import_dataset.append(ds, notify=False)
                    pn.config.notifications = notifications
                
                if self.import_dataset is not None:
                    #perror("get_data")
                    wtabulator.value = self.import_dataset.sequences
                else:
                    #perror("empty data")
                    wtabulator.value = pd.DataFrame(columns=list(dtype.keys()))
            except Exception as inst:
                logger.error(f'on_import : {inst}', exc_info=True)
            finally:
                self._layout.loading = False
                                
        
        def on_save(event):
            """ Handle saving of the imported dataset. """
            if self.import_dataset is None:
                raise ValueError('The imported dataset is empty.')
            if dataset_name.value == '':
                raise ValueError('The dataset name is empty.')      
            self._layout.loading = True
            try:
                fn = unique_filename(Path(self.dendron_info.dataset_path) / dataset_name.value)      
                self.import_dataset.dump(fn)
                #logger.info(f'Save done in dataset {fn}')
                self.dataset_selector.on_refresh(event)
            finally:
                self._layout.loading = False

        def on_merge(event):
            """ Handle saving of the imported dataset. """
            if self.import_dataset is None:
                raise ValueError('The imported dataset is empty.')
            if self.dataset is None:
                raise ValueError('The dataset is empty.')            
            self._layout.loading = True
            try:
                self.dataset.append(self.import_dataset)
                #logger.info('Merge done in current dataset')
            finally:
                self._layout.loading = False

        
        """ Create the import tab layout. """

        def on_clear_import(event):
            self.import_dataset = None
            file_input.value = {}
            wtabulator.value = pd.DataFrame(columns=list(dtype.keys()))

        def sync_columns(event):
            self.param_columns.update_tabulator(wtabulator)
            # wtabulator.hidden_columns = _hidden_columns(self.wcolumns.value)
            # if wtabulator.value is not None:
            #     wtabulator.value = wtabulator.value[_position_columns(self.wcolumns.value, wtabulator.hidden_columns)]

        option = ['Besançon', 'Heidelberg', 'RWL']
        option += ['Sylphe', 'DendronIV', 'pyDendron']
        
        file_input = pn.widgets.FileDropper(sizing_mode='stretch_both', multiple=True, max_file_size='1000MB')
        #file_input.param.watch(on_file_input, ['value'], onlychanged=True)
        
        #file_format = pn.widgets.RadioBoxGroup(name='Format', options=option, inline=True, align=('end', 'end'))
        bt_import = pn.widgets.Button(name='Import', icon='import', button_type='primary', align=('end', 'end'), description='Import the selected file.')
        bt_import.on_click(on_import)
        bt_clear_import = pn.widgets.Button(name='Clear import', icon='trash', button_type='primary', align=('end', 'end'), description='clear imported data.')
        bt_clear_import.on_click(on_clear_import)
        
        dtype = self.param_columns.get_dtype()
        wtabulator = pn.widgets.Tabulator(
                                    pd.DataFrame(columns=self.param_columns.get_columns_hiddens()),
                                    hidden_columns=self.param_columns.get_hiddens(), 
                                    text_align=_cell_text_align(dtype),
                                    editors=_cell_editors(dtype), 
                                    header_filters=_header_filters(dtype), 
                                    formatters=_cell_formatters(dtype),
                                    sizing_mode='stretch_width',
                                    frozen_columns=[ICON, KEYCODE], 
                                    selectable='checkbox',
                                    pagination='local',
                                    page_size= 1000, 
                                    max_height=600,
                                    min_height=300,
                                    height_policy='max',
                                    layout='fit_data_fill',
                                    margin=(0,0),
                                    )
        self.param_columns.get_widgets().param.watch(sync_columns, ['value'], onlychanged=True, queued=True)
        
        dataset_name = pn.widgets.TextInput(name='Dataset name', placeholder='dataset_import.p', description='Enter the dataset name.')
        bt_save = pn.widgets.Button(name='Create', icon='file-plus', button_type='primary', align=('start', 'end'), description='Save the imported dataset in a new dataset.')
        bt_save.on_click(on_save)
        bt_merge = pn.widgets.Button(name='Merge in current dataset', icon='arrow-merge', button_type='primary', align=('start', 'end'), description='Merge the imported dataset in the current dataset.')
        bt_merge.on_click(on_merge)
        return pn.Column(file_input,
                        pn.Row(bt_clear_import, bt_import ),
                        wtabulator,
                        pn.Row(dataset_name, bt_save),
                        bt_merge
                    )

    def _export(self):
        def save_dataset_to_file():
            nonlocal fn_dataset
            if self.dataset.filename is None:
                logger.warning('Dataset is not loaded.')
                return
            fn_dataset = Path(self.tmp_path) / Path(self.dataset.filename).name
            if file_export_format.value == 'Pickle':
                fn_dataset = fn_dataset.with_suffix('.p')
            elif file_export_format.value == 'JSON':
                fn_dataset = fn_dataset.with_suffix('.json')
            else:
                fn_dataset = fn_dataset.with_suffix('.xlsx')
            fn_dataset = unique_filename(fn_dataset)
            #logger.info(f'Tools export, filename: {fn_dataset} (format {file_export_format.value})')
            self.dataset.dump(fn_dataset)
            #self.dataset_selector.on_refresh(event)
        
        def on_export_dataset():
            nonlocal fn_dataset
            try:
                save_dataset_to_file()
                bt_export_dataset.filename = fn_dataset.name
                output = BytesIO()
                with open(fn_dataset, 'rb') as file:
                    output.write(file.read())
                output.seek(0)
                fn_dataset.unlink()
                return output
            except Exception as inst:
                logger.error(f'on_excel: {inst}', exc_info=True)
            
        def on_export_package():
            nonlocal fn_package
            try:
                save_package_to_file()
                bt_export_package.filename = fn_package.name
                output = BytesIO()
                with open(fn_package, 'rb') as file:
                    output.write(file.read())
                output.seek(0)
                fn_package.unlink()
                return output
            except Exception as inst:
                logger.error(f'on_excel: {inst}', exc_info=True)
    
        def save_package_to_file():
            nonlocal fn_package
            if self.dataset.filename is None:
                logger.warning('Dataset is not loaded.')
                return
            data = dataset_package.data
            if data is None:
                logger.warning('No package to export.')
                return
            package = dataset_package.get_package_name().replace(os.sep, '_')
            if file_export_format_package.value == 'Besançon':
                bes = IOBesancon()
                fn_package = Path(self.tmp_path) / f'{package}.bes'
                bes.write_package(self.dataset, data, fn_package)
                #logger.info(f'Tools export {package}, filename: {fn} (format {file_export_format_package.value})')
            elif file_export_format_package.value == 'Heidelberg':
                fh = IOHeidelberg()
                fn_package = Path(self.tmp_path) / f'{package}.fh'
                fh.write_package(self.dataset, data, fn_package)
                #logger.info(f'Tools export {package}, filename: {fn} (format {file_export_format_package.value})')
            # elif file_export_format_package.value == 'RWL':
            #     rwl = IORWL()
            #     fn = Path(self.path) / f'{package}.rwl'
            #     rwl.write_package(self.dataset, data, fn)
            #     #logger.info(f'Tools export {package}, filename: {fn} (format {file_export_format_package.value})')
            elif file_export_format_package.value == 'Excel Rows':
                fn_package = Path(self.tmp_path) / f'{package}_row.xlsx'
                with pd.ExcelWriter(fn_package) as writer:
                    data.to_excel(writer, sheet_name=package, merge_cells=False, float_format="%.6f")
                #logger.info(f'Tools export {package}, filename: {fn} (format {file_export_format_package.value})')
            elif file_export_format_package.value == 'Excel Columns':
                fn_package = Path(self.tmp_path) / f'{package}_col.xlsx'
                if data[DATE_BEGIN].notna().any():
                    #perror('data2col date_begin')
                    data_col = data2col(data, use_offset=True, key_offset=DATE_BEGIN, add_index=True)
                    data_col_w = data2col(data, value_key=DATA_WEIGHTS , use_offset=True, key_offset=DATE_BEGIN, raise_error=False, add_index=True)
                    data_col_s = data2col(data, value_key=DATA_SIGNATURES, use_offset=True, key_offset=DATE_BEGIN, raise_error=False, add_index=True)
                elif data[OFFSET].notna().any():
                    #perror('data2col offset')
                    data_col = data2col(data, use_offset=True, key_offset=OFFSET, add_index=True)
                    data_col_w = data2col(data, value_key=DATA_WEIGHTS, use_offset=True, key_offset=OFFSET, raise_error=False, add_index=True)
                    data_col_s = data2col(data, value_key=DATA_SIGNATURES,  use_offset=True, key_offset=OFFSET, raise_error=False, add_index=True)
                else:
                    #perror('data2col ')
                    data_col = data2col(data, use_offset=False, add_index=True)
                    data_col_w = data2col(data, value_key=DATA_WEIGHTS, use_offset=False, raise_error=False, add_index=True)
                    data_col_s = data2col(data, value_key=DATA_SIGNATURES, use_offset=False, raise_error=False, add_index=True)

                with pd.ExcelWriter(fn_package) as writer:
                    data[excel_columns].T.to_excel(writer, sheet_name='Info', merge_cells=False, float_format="%.6f")
                    data_col.to_excel(writer, sheet_name=DATA_VALUES, merge_cells=False, float_format="%.6f")
                    data_col_w.to_excel(writer, sheet_name=DATA_WEIGHTS, merge_cells=False, float_format="%.6f")
                    data_col_s.to_excel(writer, sheet_name=DATA_SIGNATURES, merge_cells=False, float_format="%.6f")
                #logger.info(f'Tools export {package}, filename: {fn} (format {file_export_format_package.value})')
                    
                #self.dataset_selector.on_refresh(event)
            else:
                logger.warning(f'Unknown format {file_export_format_package.value}')
                return
             
        def on_stats(event):
            try:
                dataset_package.add_statistics = cb_stats.value
                perror(f'Add statistics {cb_stats.value}')
            except Exception as inst:
                logger.error(f'on_excel: {inst}', exc_info=True)     
             
             
                 
        file_export_format = pn.widgets.RadioBoxGroup(name='file format', options=['Pickle', 'Excel', 'JSON'], inline=True, align=('start', 'end'))

        bt_export_dataset = pn.widgets.FileDownload(callback=on_export_dataset, filename='dataset.tmp', embed=False, 
                                                label='Download', icon='file-export', button_type='primary', 
                                                width=self.bt_size, align=('start', 'end'), description='Download exported package.')
        
        #bt_save_dataset.on_click(save_dataset_to_file)
        
        file_export_format_package = pn.widgets.RadioBoxGroup(name='file format', options=['Heidelberg', 'Besançon', 'Excel Columns', 'Excel Rows'], inline=True, align=('start', 'end'))
        #bt_save_dataset_package = pn.widgets.Button(name='Export', icon='file-arrow-right', button_type='primary', align=('start', 'end'), width=self.bt_size, description='Export the selected dataset.')
        #bt_save_dataset_package.on_click(save_package_to_file)
        fn_package = ''
        fn_dataset = ''
        
        cb_stats = pn.widgets.Checkbox(name='Add statistics', align=('start', 'end'))
        cb_stats.param.watch(on_stats,  ['value'], onlychanged=True)

        
        bt_export_package = pn.widgets.FileDownload(callback=on_export_package, filename='package.tmp', embed=False, 
                                                label='Download', icon='file-export', button_type='primary', 
                                                width=self.bt_size, align=('start', 'end'), description='Download exported package.')
        
        dataset_package = self.dataset_package
        return pn.Column(
                    pn.pane.Markdown('### Export current dataset'),
                    file_export_format,
                    bt_export_dataset,
                    pn.pane.Markdown('### Export current package'),
                    dataset_package,
                    cb_stats,
                    file_export_format_package,
                    #bt_save_dataset_package,
                    bt_export_package,
        )
  
    def _merge(self):
        """ Create the dataset merge tab layout. """
        
        def load(dataset_src_name1, dataset_src_name2):
            dataset_src = Dataset()
            dataset_src.load(dataset_src_name1.value)
            
            dataset_dest = Dataset()
            dataset_dest.load(dataset_src_name2.value)
            
            return  dataset_src, dataset_dest

        def on_merge(event):
            """ Handle merging of datasets. """
            try:
                self._layout.loading = True
                fn = unique_filename(Path(self.dendron_info.dataset_path) / dataset_dest_name.value)
                #perror(f'Merge {dataset_src_name1.value} and {dataset_src_name2.value} in {fn}')
                if dataset_src_name1.value == dataset_src_name2.value:
                    logger.warning('The source datasets are the same.')
                    return
                dataset_src1, dataset_src2 = load(dataset_src_name1, dataset_src_name2)
                dest = dataset_src1.clone()
                dest.filename = fn
                dest.append(dataset_src2)
                dest.dump()
                #logger.info(f'Merge done into {fn}')
            except Exception as inst:
                logger.error(f'on_duplicate2 : {inst}', exc_info=True)
            finally:
                self._layout.loading = False

        options = self.get_options()
        
        dataset_src_name1 = pn.widgets.Select(name='1st dataset name', options=options, description='Select the first dataset to merge.')
        dataset_src_name2 = pn.widgets.Select(name='2rd dataset name', options=options, description='Select the second dataset to merge.')
        dataset_dest_name = pn.widgets.TextInput(value='dataset_import.p', name='Destination Dataset name', description='Enter the destination dataset name.')

        bt_merge = pn.widgets.Button(name='Merge', icon='arrow-merge', button_type='primary', align=('start', 'end'), width=self.bt_size, description='Merge the 2 datasets.')
        bt_merge.on_click(on_merge)

        return pn.Column(
                        pn.Row(dataset_src_name1, dataset_src_name2), 
                        pn.Row(dataset_dest_name, bt_merge)
                    )
        
    def _validate(self):

        parent_orphans = []
        child_orphans = []
        seq_orphans = []
                
        def on_duplicate(event):
            try:
                self._layout.loading = True
                counts = self.dataset.sequences[KEYCODE].value_counts()
                lst = list(counts[counts > 1].to_dict().keys())
                wduplicate.value =  '\n'.join(lst) if len(lst) > 0 else 'No duplicate'
            except Exception as inst:
                logger.error(f'on_duplicate2 : {inst}', exc_info=True)
            finally:
                self._layout.loading = False

        def on_duplicate2(event):
            try:
                self._layout.loading = True
                data = dataset_package.data
                if data is None:
                    logger.warning('No package.')
                    return
                res = alignment.ndiff_sequences(data, n=self.tolerance)
                txt = ''
                for id_i, id_j, cost , insertions, deletions, substitutions, n, err in res:
                    #id_i, id_j, cost , insertions, deletions, substitutions, n, ch]
                    txt += f'{err} (ins: {insertions}, del:{deletions}, sub:{substitutions})\n'
                wduplicate2.value = txt
                progress.value = 100
            except Exception as inst:
                logger.error(f'on_duplicate2 : {inst}', exc_info=True)
            finally:
                self._layout.loading = False
        
        def on_progress(event):
            progress.value = alignment.rate
            
        def on_outliers(event):
            data = dataset_package.dt_data
            if data is None:
                logger.warning('No package.')
                return
            txt = ''
            self.outliers.clear()
            
            for _, row in data.iterrows():
                vec = row[DATA_VALUES]
                id = row[ID_CHILD]
                med, std = np.nanmedian(vec), np.nanstd(vec)
                cut_off = std * 5 
                lower, upper = med - cut_off, med + cut_off
                keycode = row[KEYCODE]
                for i, x in enumerate(vec):
                    if (x < lower) or (x > upper) or (abs(x) <= 1e-7):
                        txt += f'Outlier {x} in "{keycode}" at index {i} (lower={round(lower, 3)}, median={round(med, 3)}, upper={round(upper, 3)})\n'
                        self.outliers.append(({ID: id, KEYCODE: keycode, 'Index':i, 'Value': x}))
            woutliers.value = txt

        def on_set_nan(event):
            for d in self.outliers:
                self.dataset.sequences.at[d[ID], DATA_VALUES][d['Index']] = np.nan
            self.dataset.notify_changes('outliers')
            
        def on_orphans(event):
            nonlocal parent_orphans, child_orphans, seq_orphans
            try:        
                self._layout.loading = True
                parent_orphans.clear()
                child_orphans.clear()
                seq_orphans.clear()
                parent_orphans += self.dataset.get_orphans_components(level=ID_PARENT)
                child_orphans += self.dataset.get_orphans_components(level=ID_CHILD)
                seq_orphans += self.dataset.get_orphans_sequences()
                worphans.value = f'Components {ID_PARENT}:\n{parent_orphans}\n Components {ID_CHILD}:\n{child_orphans}\n Sequences:\n{seq_orphans}'
            except Exception as inst:
                logger.error(f'on_orphans : {inst}', exc_info=True)
            finally:
                self._layout.loading = False
        
        def on_delete_orphans(event):
            try:
                self._layout.loading = True
                if len(parent_orphans) > 0:
                    self.dataset.drop_orphans_components(parent_orphans)
                if len(child_orphans) > 0:
                    self.dataset.drop_orphans_components(child_orphans)
                on_orphans(event)
            except Exception as inst:
                logger.error(f'on_delete_orphans : {inst}', exc_info=True)
            finally:
                self._layout.loading = False
        
        def on_check(event):
            try:
                self._layout.loading = True
                problems = []
                for id, row in self.dataset.sequences.iterrows():
                    keycode = row[KEYCODE]
                    data = row[DATA_VALUES]
                    #perror(f'Check {keycode} {data}')
                    if isinstance(data, np.ndarray):
                        ldata = len(data)
                        if ldata != row[DATA_LENGTH]:
                            problems.append(f'In {keycode}, length {len(data)} different from {row[DATA_LENGTH]}.')
                        if pd.notna(row[SAPWOOD] ):
                            if row[SAPWOOD] >= ldata :
                                problems.append(f'In {keycode}, {SAPWOOD} value {row[SAPWOOD]} is greater than data lenght.')
                            elif row[SAPWOOD] < 0 :
                                problems.append(f'In {keycode}, {SAPWOOD} value {row[SAPWOOD]} is negative.')
                    if pd.notna(row[DATE_BEGIN]) and pd.notna(row[DATE_END]):
                        if row[DATE_BEGIN] > row[DATE_END]:
                            problems.append(f'In {keycode}, {DATE_BEGIN} {row[DATE_BEGIN]} is greater than {DATE_END}')
                        if isinstance(data, np.ndarray):
                            ldata = len(data)
                            if (row[DATE_END] - row[DATE_BEGIN] + 1) != ldata:
                                problems.append(f'In {keycode}, {DATE_END} - {DATE_BEGIN} + 1 = {row[DATE_END] - row[DATE_BEGIN] + 1} different from data length {ldata}.')                       
                wcheck.value = '\n'.join(problems) if len(problems) > 0 else 'No problem detected.'
            except Exception as inst:
                logger.error(f'on_check : {inst}', exc_info=True)
            finally:
                self._layout.loading = False
        
        dataset_package = self.dataset_package
        bt_duplicate = pn.widgets.Button(name=f'Detect duplicate {KEYCODE}', icon='bolt', button_type='primary', description='Detect duplicate in the current dataset.')
        bt_duplicate.on_click(on_duplicate)
        wduplicate = pn.widgets.TextAreaInput(name=f'Duplicate {KEYCODE}', value='', sizing_mode='stretch_width', description='List of duplicate keycodes.')

        alignment = Alignment()
        bt_duplicate2 = pn.widgets.Button(name=f'Detect duplicate {DATA_VALUES}', icon='bolt', button_type='primary', description=f'Detect duplicate {DATA_VALUES} in the current package. The tolerance is set to {self.tolerance}.')
        bt_duplicate2.on_click(on_duplicate2)
        progress = pn.indicators.Progress(name='Run', value=0, sizing_mode='stretch_width', disabled=True, bar_color='primary')
        alignment.param.watch(on_progress, ['rate'], onlychanged=True)
        wduplicate2 = pn.widgets.TextAreaInput(name=f'Duplicate {DATA_VALUES}', value='', sizing_mode='stretch_width', description=f'List of {KEYCODE} with duplicate values.')

        bt_outliers = pn.widgets.Button(name=f'Detect outliers {DATA_VALUES}', icon='bolt', button_type='primary', description=f'Detect outliers in the current package. Outliers are values outside 5 standard deviations from the median.')
        bt_outliers.on_click(on_outliers)
        woutliers = pn.widgets.TextAreaInput(name=f'Outliers {DATA_VALUES}', value='', sizing_mode='stretch_width', description='List of {KEYCODE} with outlier values.')
        bt_set_nan = pn.widgets.Button(name='Set value to NaN', icon='bolt', button_type='primary', description='Set the outlier values to NaN.')
        bt_set_nan.on_click(on_set_nan)

        bt_orphans = pn.widgets.Button(name='Detect', icon='bolt', button_type='primary', description='Detect orphan indexes in database.')
        bt_orphans.on_click(on_orphans)
        worphans = pn.widgets.TextAreaInput(name=f'{KEYCODE} orphans', value='', sizing_mode='stretch_width', description='List of detected orphans.')
        bt_delete_orphans = pn.widgets.Button(name='Delete', icon='delete', button_type='primary', description='Delete detected orphans.')
        bt_delete_orphans.on_click(on_delete_orphans)

        bt_check = pn.widgets.Button(name='Check', icon='bolt', button_type='primary', description=f'Check {DATA_LENGTH}, {SAPWOOD}, date.')
        wcheck = pn.widgets.TextAreaInput(name=f'Problems', value='', sizing_mode='stretch_width', description='List of detected problems.')
        bt_check.on_click(on_check)

        return pn.Tabs(
            (f'Duplicate {KEYCODE}', pn.Column(bt_duplicate, wduplicate)),
            (f'Duplicate {DATA_VALUES}', pn.Column(dataset_package, pn.Row(bt_duplicate2,  progress), wduplicate2)),
            (f'Outliers {DATA_VALUES}', pn.Column(dataset_package, bt_outliers, woutliers, bt_set_nan)),
            ('Orphans', pn.Column(bt_orphans, worphans, bt_delete_orphans)),
            ('Check data', pn.Column(bt_check, wcheck)),
            dynamic=True
        )

    def _edit_ring(self):
        def on_insert(event):
            if len(wtabulator.selection) > 0:
                df = wtabulator.value
                i = wtabulator.selection[0]
                with warnings.catch_warnings():
                    # TODO: pandas 2.1.0 has a FutureWarning for concatenating DataFrames with Null entries
                    warnings.filterwarnings("ignore", category=FutureWarning)
                    wtabulator.value = pd.concat([df.iloc[:i], pd.DataFrame({'value' : [pd.NA]}), df.iloc[i:]]).reset_index(drop=True)
        
        def on_append(event):
            with warnings.catch_warnings():
                # TODO: pandas 2.1.0 has a FutureWarning for concatenating DataFrames with Null entries
                warnings.filterwarnings("ignore", category=FutureWarning)
                wtabulator.value = pd.concat([wtabulator.value, pd.DataFrame({'value' : [pd.NA]})]).reset_index(drop=True)

        def on_delete(event):
            if len(wtabulator.selection) > 0:
                wtabulator.value = wtabulator.value.drop(wtabulator.selection).reset_index(drop=True)
                wtabulator.selection = []
        
        def on_save(event):
            try:
                self._layout.loading = True
                id = select.value
                #perror(f'on_save {id}')
                array = wtabulator.value['value'].to_numpy()
                self.dataset_package.dataset.set_inconsistent_on_ascendants(id, notify=False)           
                self.dataset_package.dataset.edit_sequence(id, DATA_VALUES, array)     
                
                """
                id = select.value
                array = wtabulator.value['value'].to_numpy()
                old_len = dataset_package.dataset.sequences.at[id, DATA_LENGTH]
                length = len(array)
                if length != old_len :
                    self.dataset_package.dataset.edit_sequence(id, DATA_LENGTH, length, notify=False)                
                    self.dataset_package.dataset.edit_sequence(id, DATA_LENGTH, dataset_package.dataset.sequences.at[id, DATE_BEGIN] + length, notify=False)                
                ascendants = self.dataset_package.dataset.get_ascendants(id, recursive=True)
                if len(ascendants) > 0:
                    self.dataset_package.dataset.edit_sequence(ascendants, INCONSISTENT, True, notify=False)
                self.dataset_package.dataset.edit_sequence(id, DATA_VALUES, array)                
                wtabulator.value """
            except Exception as inst:
                logger.error(f'on_save : {inst}', exc_info=True)
            finally:
                self._layout.loading = False

        def sync_data(event):
            options = {}
            if dataset_package.dt_data is not None :
                for i, row in dataset_package.dt_data.iterrows():
                    options[row[KEYCODE]] = row[ID_CHILD]
            select.options = options
            select.value = None

        def sync_tabulator(event):
            if select.value is not None:
                category = dataset_package.dataset.sequences.at[select.value, CATEGORY]
                if category == MEASURE:
                    data = dataset_package.dataset.sequences.loc[select.value, DATA_VALUES]
                    wtabulator.value = pd.DataFrame({'value' : data})
                else:
                    wtabulator.value = None
                    logger.warning(f'The selected serie is not a tree ring serie.')

        dataset_package = self.dataset_package
        dataset_package.wselection.param.watch(sync_data, ['value'], onlychanged=True)
        #dataset_package.panel_tabulator.visible = False
        
        select = pn.widgets.Select(name=f'Select {KEYCODE}', options=[], description=f'Select the {KEYCODE} serie to edit.')
        select.param.watch(sync_tabulator, ['value'], onlychanged=True)
        
        bt_insert = pn.widgets.Button(name=f'Insert', icon='insert', button_type='primary', description='Insert a new value before the selected row.')
        bt_insert.on_click(on_insert)

        bt_append = pn.widgets.Button(name=f'Append', icon='append', button_type='primary', description='Append a new value after the last row.')
        bt_append.on_click(on_append)

        bt_delete = pn.widgets.Button(name=f'Delete', icon='delete', button_type='primary', description='Delete the selected rows.')
        bt_delete.on_click(on_delete)

        bt_save = pn.widgets.Button(name=f'Save', icon='save', button_type='primary', description='Save the modifications.')
        bt_save.on_click(on_save)

        wtabulator = pn.widgets.Tabulator(
                                    sizing_mode='stretch_width',
                                    max_height=600,
                                    min_height=300,
                                    height_policy='max',
                                    #selectable=True,
                                    selectable='checkbox',

                                    show_index = True
                                    )
        
        return pn.Column(
            dataset_package,
            select,
            wtabulator,
            pn.Row(bt_insert, bt_append, bt_delete, bt_save)
        )
     
    def __panel__(self):
        """ Return the panel layout."""
        return self._layout
    



        






"""
    Import Syphe database in CSV format
"""
__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"




import logging
from pathlib import Path
import datetime
import pickle

import numpy as np 
import pandas as pd

from pyDendron.dataset import Dataset
from pyDendron.app_logger import logger, perror
from pyDendron.dataname import *
from pyDendron.alien.mdb2csv import read_table

from pyDendron.tools.location import add_geocode

sylphe_col_sequences = {   
    'NumSeq': ID, #int
    'libellé': KEYCODE, #string  

    'espèce':SPECIES, #string
    #'NatureSeq' : 'object.type', #string
    #'Forme': 'object.shape', # string

    'NumGeo': SITE_CODE, #string
    #'DouteGeo' :'place.doubt', # bool

    'origine': DATE_BEGIN, #float
    'terme' : DATE_END, #float
    #'termeOptimum' : DATE_END_OPTIMUM, #float
    #'termeMaxi' : DATE_END_MAXIMUM, #float
    'DateSupose' : SYNC, # bool

    'NumTypseq' : CATEGORY, #string
    'NumSousTypSeq' : SUBCATEGORY, #string

    #'Aubier' : SAPWOOD, # bool
    'PosAubier' : SAPWOOD, #float

    'Moelle' : PITH, # bool
    'PosMoellerEstime' : PITH_ESTIMATED, #float

    'Cambium' : CAMBIUM, # bool
    #'PosCambiumEstimeMaxi' : CAMBIUM_MAXIMUM, #float
    #'PosCambiumEstimeOpti': CAMBIUM_OPTIMUM, #float
    #'typeEstimation': CAMBIUM_METHOD, #float
    'CambiumType' : CAMBIUM_SEASON, #string

    'Ecorce' : BARK, # bool

    'DateCréation': CREATION_DATE, # date ?, #string
    #'Informateur' : 'creation.owner', # string
    'NumPersonne' : PERS_ID, # string
    'Note': COMMENTS # string
    }

sylphe_col_sequences_miss = [
        DATA_LENGTH, DATA_VALUES, DATA_TYPE, DATA_WEIGHTS,
        ]

sylphe_col_components = {'NumMean' : ID_PARENT, 'NumSeq' : ID_CHILD, 'Pos' : OFFSET}
#sylphe_col_components_miss = ['keycode.child', 'path.child', ]
sylphe_col_indices = {'NumSeq' : ID, 'Longueur': 'count', 'TypeInd' : 'key', 'Val' : 'values'}
sylphe_col_indices_miss = []

class IOSylphe:
    """
    Class to import Sylphe database after conversion in CSV files with '@' as separator
    """
    file_extension = ['.mdb']

    def __init__(self, base_directory='./',get_place=False, places='location.p', reverse_places='reverse_location.p'):
        self.get_place = get_place
        self.places = {}
        self.places_fn = places
        self.reverse_places_fn = reverse_places
        self.elevations = {}
        self.base_directory = base_directory

        self.sequences = None
        self.components = None
        
    def read_sequences(self, sep : str ='@', header : int =0) :
        """
        Read cvs sequences

        Keyword Arguments:
            sep -- fields separator (default: {'@'})
            header -- head column values (default: {0})
            usecols -- columns to keep (default: 
            rename_columns -- how to rename columns 
        Returns:
            a Pandas DataFrame and a list sequences with trouble
        """

        def _places():
            def load(name):
                if (name is not None) and Path(name).exists():
                    with open(name , 'rb') as fic:
                        dic = pickle.load(fic)
                else:
                    dic = {}
                return dic
            def load2(name):
                if (name is not None) and Path(name).exists():
                    with open(name , 'rb') as fic:
                        dic = pickle.load(fic)
                else:
                    dic = [{}, {}]
                return dic
            places = load(self.places_fn)
            reverse_places, elevation = load2(self.reverse_places_fn)
            
            places, reverse_places, elevation = add_geocode(sequences, places, reverse_places, elevation)
            with open(self.places_fn , 'wb') as fic:
                pickle.dump(places, fic)
                
            with open(self.reverse_places_fn , 'wb') as fic:
                pickle.dump([reverse_places, elevation], fic)

        def _gregorian():
            # set gregorian date to astronomique gregorian date
            for key in [DATE_BEGIN, DATE_END]: #, DATE_END_OPTIMUM, DATE_END_MAXIMUM]:
                sequences[key] = sequences[key].astype('Int32')
                sequences.replace({key : {0: pd.NA}}, inplace=True) 
                sequences.loc[sequences[key] < 0, key] += 1

        def _type():
            for key in [SAPWOOD]: 
                sequences[key] = sequences[key].apply(lambda x: pd.NA if (x > 159394000) or (x == 0) or (x == 65535) else x)
                sequences[key] = sequences[key].astype('Int32')
                sequences.loc[sequences[key] > 0, key] -= 1

            for key in [BARK, CAMBIUM, PITH]:
                sequences[key] = sequences[key] > 0
                sequences[key] = sequences[key].astype(bool)


        #filename = Path(self.path) / Path(self.sequences_filename)
        #sequences = pd.read_csv(filename, sep=sep, header=header, usecols=sylphe_col_sequences.keys())
        sequences = read_table(self.mdb_file, 'Sequence', converters_from_schema=False, usecols=sylphe_col_sequences.keys()) 
        
        
        sequences.rename(columns=sylphe_col_sequences, inplace=True)
        sequences = sequences[sylphe_col_sequences.values()]

        for col in sylphe_col_sequences_miss:
            sequences[col] = None

        sequences.replace({CATEGORY: {'MEAN': MEAN, 'SET': SET, 'SEQ': MEASURE}}, inplace=True)

        _gregorian()
        _type()
        if self.get_place: _places()
        
        self.sequences = sequences.set_index(ID)
        
        self._check_sequences()
        
        return self.sequences

    def _string(self, dataframe, key):
        dataframe[key] = dataframe[key].astype('string')

    def _check_sequences(self):
        # Check lastYear and firstYear consistency
        year = datetime.datetime.now().year
        date1 = self.sequences.loc[(self.sequences[DATE_END] - self.sequences[DATE_BEGIN]) < 0].index.tolist()
        date2 = self.sequences.loc[(self.sequences[DATE_END] > year) | 
                                (self.sequences[DATE_BEGIN] > year)].index.tolist()

        trouble = list()
        if date1 is not None:
            trouble = date1
        if date2 is not None:
            trouble += date2
        if len(trouble) > 0:
            for id in trouble:
                row = self.sequences.loc[id]
                logger.warning(f'inconsistent dates, to ckeck: {id} {row[KEYCODE]} (begin: {row[DATE_BEGIN]}, end:{row[DATE_END]})')

        return trouble

    def read_indices(self,sep='@', header=0):
        """Read CSV Indices

        Keyword Arguments:
            sep -- fields separator (default: {'@'})
            header -- head column values (default: {0})
        Returns:
            a indices object and a list sequences with trouble
        """
        def count_nan_in_vector(vector):
            if vector is None:
                return 0
            return np.isnan(vector).sum()
        
        def _read_values():
            values = []
            weights = []
            tmp = row['values'].replace(',', '.')
            tmp = tmp.replace('\\r\\n', '')
            for line in tmp.split(';')[:-1]:
                try :
                    s = line.split()
                    x = s[1]
                    w = s[3]
                    x = np.nan if x == 'ABS' else float(x)
                    x = np.nan if (x == 999) or (x == 9999) else x
                except ValueError:
                    x = np.nan
                    w = 0
                    logger.warning(f'error in values {id} {self.sequences.loc[id, KEYCODE]} | {line}')
                weights.append(w)
                values.append(x)
            
            if len(values) == 0:
                return pd.NA, 0, pd.NA
            
            return np.array(values), len(values), np.array(weights)
        
        def _check_dates(length):
            keycode, begin, end = self.sequences.loc[id, [KEYCODE, DATE_BEGIN, DATE_END]]
            c = row['count']
            s = f'end-begin dates != values length for: {id} {keycode} : len != (DateEnd - DateBegin + 1)  {length} {c} != ( {end} - {begin} + 1 = {end - begin + 1})',  
            if not pd.isna(end) and not pd.isna(begin) and length != (end - begin + 1):
                trouble.append([id, length, begin, end])
                logger.warning(s)

        # -------
        #fn = Path(self.path) / Path(self.indices_filename)
        #indices = pd.read_csv(fn, sep=sep, header=header, usecols=sylphe_col_indices.keys())
        indices = read_table(self.mdb_file, 'Indice', converters_from_schema=False, usecols=sylphe_col_indices.keys()) 

        indices.rename(columns=sylphe_col_indices, inplace=True)

        # Get raw indices
        indices = indices.loc[(indices['key'] == 'NAT') & (indices[ID].isin(self.sequences.index)), :]
        
        trouble = []
        to_drop = []
        for id, df in indices.groupby(ID): 
            row = df.iloc[0]
            values, length, weights = _read_values()
            self.sequences.at[id, DATA_VALUES] = values
            self.sequences.at[id, DATA_LENGTH] = length
            self.sequences.at[id, DATA_TYPE] = RAW

            if self.sequences.at[id, CATEGORY] == MEAN:
                self.sequences.at[id, DATA_WEIGHTS] = weights
            _check_dates(length)

            if (length == count_nan_in_vector(values)):
                if self.sequences.at[id, CATEGORY] == MEASURE:
                    to_drop.append(id)
                else:
                    self.sequences.at[id, DATA_VALUES] = pd.NA
                    self.sequences.at[id, DATA_LENGTH] = 0
                    self.sequences.at[id, CATEGORY] = SET
                    self.sequences.at[id, DATA_WEIGHTS] = pd.NA
        
        if len(to_drop) > 0:
            logger.warning(f'Drop empty or full NaN {DATA_VALUES} for samples:  {to_drop}')
            self.sequences.drop(to_drop, inplace=True)
        
        for k, t in sequences_dtype_dict.items():
            if k not in self.sequences.columns:
                self.sequences[k] = ''
            if t == 'string':
                self.sequences[k] = self.sequences[k].fillna('')
        
        self.sequences[SPECIES] = self.sequences[SPECIES].str.upper()
        mask = self.sequences[SPECIES].str.startswith('QU')
        self.sequences.loc[mask, SPECIES] = 'QU'
        
        self.sequences[SUBCATEGORY] = self.sequences[SUBCATEGORY].str.upper()
        self.sequences[COMMENTS] = self.sequences[COMMENTS].str.replace(r'\n', '.')
        self.sequences[COMMENTS] = self.sequences[COMMENTS].str.replace(r'\r', '')
        self.sequences[INCONSISTENT] = False
        
    def read_components(self, sep='@', header=0):
        """_summary_

        Keyword Arguments:
            sep -- fields separator (default: {'@'})
            header -- head column values (default: {0})
        Returns:
            a Pandas DataFrame and 3 lists with trouble
        """

        # ----------
        #filename = Path(self.path) / Path(self.components_filename)
        #components = pd.read_csv(filename, sep=sep, header=header, usecols=sylphe_col_components.keys())
        components = read_table(self.mdb_file, 'Composant', converters_from_schema=False, usecols=sylphe_col_components.keys()) 

        components.rename(columns=sylphe_col_components, inplace=True)

        self._add_missing_components_col(components)
        key = OFFSET
        components[key] = components[key].apply(lambda x: pd.NA if (x > 159394000) or (x == 0) or (x == 65535) else x)
        components[key] = components[key].astype('Int32')
        components.loc[components[key] > 0, key] -= 1
        
        components_parents = self._id_parent_is_in_sequences(components)
        self._id_child_is_in_sequences(components, components_parents)
        
        components = self._check_components(components)

        self.components = components
        return self.components

    def _add_missing_components_col(self, components):
        pass
        #for col in sylphe_col_components_miss:
        #    components[col] = None

    def _id_parent_is_in_sequences(self, components):
        owner_before = set(components[ID_PARENT].to_list())
        components_parents = components.loc[components[ID_PARENT].isin(self.sequences.index)]
        owner_after = set(components[ID_PARENT].to_list())
        owner_removed = owner_before - owner_after
        if len(owner_removed) > 0:
            logger.warning(f'remove components without sequence:  {owner_removed}')
        return components_parents
        
    def _id_child_is_in_sequences(self, components, components_parents):
        id_before = set(components[ID_CHILD].to_list())
        components = components_parents.loc[components[ID_CHILD].isin(self.sequences.index)]
        id_after = set(components[ID_CHILD].to_list())
        id_removed = id_before - id_after

        if len(id_removed) > 0:
            logger.warning(f'remove components without sequence:  {id_removed}')
            
    def _check_components(self, components):
        components_with_date_begin = components.merge(self.sequences[DATE_BEGIN], left_on=ID_CHILD, right_on=ID)
        
        for id_parent, data in components_with_date_begin.groupby(ID_PARENT):
            min_date = data[DATE_BEGIN].min()
            min_pos = data[OFFSET].min()
            if pd.isna(min_date) or pd.isna(min_pos):
                components_with_date_begin.loc[components_with_date_begin[ID_PARENT] == id_parent, DATE_BEGIN] = 0
                components_with_date_begin.loc[components_with_date_begin[ID_PARENT] == id_parent, OFFSET] = 0
            else:
                components_with_date_begin.loc[components_with_date_begin[ID_PARENT] == id_parent, DATE_BEGIN] -= min_date
                components_with_date_begin.loc[components_with_date_begin[ID_PARENT] == id_parent, OFFSET] -= min_pos
        
        mask = (components_with_date_begin[OFFSET] != components_with_date_begin[DATE_BEGIN])
        components_with_date_begin.loc[mask, OFFSET] = components_with_date_begin.loc[mask, DATE_BEGIN]
        return components_with_date_begin.drop(DATE_BEGIN, axis=1)
    
    def set_project(self):
        components = self.dataset.components
        leaf = self.dataset.get_leafs()
        for id in leaf:
            filter = components.index.get_level_values(ID_CHILD) == id
            if np.sum(filter) == 1:
                id_parent = components[filter].index.get_level_values(ID_PARENT).to_list()[0]
                self.dataset.sequences.at[id, PROJECT] = self.dataset.sequences.at[id_parent, KEYCODE]
        
    def to_dataset(self, keycode_parent='Dataset',root_id=ROOT, trash_keycode='Trash', clipboard_keycode='Clipboard', workshop_keycode='Workshop' ):
        logger.info('read sequences')
        self.read_sequences()
        logger.info('read indices')
        self.read_indices()
        logger.info('read components')
        self.read_components()
        #print(self.sequences)
        self.dataset = Dataset(self.sequences, self.components, save_auto=False)
        self.set_project()
        
        logger.info(f'to_dataset roots: {self.dataset.get_roots()}')
        
        self.dataset.new_root(keycode_parent, root_id)
        self.dataset.new_clipboard(clipboard_keycode)
        self.dataset.new_trash(trash_keycode)
        self.dataset.new_workshop(workshop_keycode)
        
        return self.dataset

    def read(self, ndb_file, keycode_parent='Sylphe'):
        self.mdb_file = Path(ndb_file)

        return self.to_dataset(keycode_parent=keycode_parent)
    
    def read_buffer(self, keycode_parent, buffer, mine_type=None):
        
        output_file = Path(self.base_directory) / Path(keycode_parent).with_suffix(self.file_extension[0])
        with open(output_file, 'wb') as f:
            f.write(buffer)  # Récupère le contenu binaire du buffer
        
        
        dataset = self.read(output_file, keycode_parent=keycode_parent) 
        output_file.unlink()
        return dataset             



"""
Mean
"""
__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"


import logging
import numpy as np
import pandas as pd
try:
    from concurrent.futures import ProcessPoolExecutor
except ImportError:
    ProcessPoolExecutor = None

from pyDendron.app_logger import logger, perror
from pyDendron.dataname import *
from pyDendron.detrend import signatures

epsilon = np.finfo(np.float64).tiny
biweight_factor = 9
#biweight_factor = 5

def data2col(data, key=ID, value_key=DATA_VALUES, use_offset=True, key_offset=OFFSET, raise_error=True, add_index=False):
    
    def length_(values):
        l = len(values) if values is not None else 0
        #logger.debug(f'{l}, {values}')
        return l
    
    data = data.copy()
    #data[DATA_LENGTH] = data[value_key].apply(lambda x: length_(x))
    #print('data2col : use_offset', use_offset)
    if use_offset:
        if data[key_offset].isna().any():
            raise ValueError(f'data2col: NA value(s) in {key_offset} column')
        begin = data[key_offset].min()
        data[key_offset] -= begin
        end = (data[key_offset]+data[DATA_LENGTH]).max()
    else:
        begin = 0
        end = data[DATA_LENGTH].max()

    d = {key_offset: np.arange(begin, end+begin)} if add_index else {}
    for id, row in data.iterrows():
        offset = row[key_offset] if use_offset else 0
        if value_key == DATA_WEIGHTS:
            values = row[value_key] if (row[value_key] is not None) else np.ones(row[DATA_LENGTH])
            #perror(f'{id} / {row[KEYCODE]}, ring {values} {row[DATA_LENGTH]}')
        else:
            values = row[value_key]
        if (values is not None) and (len(values) > 0):
            vec = np.full(end, np.nan)
            vec[offset:offset+len(values)] = values
            k = id if key == ID else row[key]
            d[k] = vec
        else:
            if raise_error:
                raise ValueError(f'data2col: {id} / {row[KEYCODE]}, ring {value_key} is missing')
            else:
                if value_key != DATA_WEIGHTS:
                    logger.warning(f'data2col: {id} / {row[KEYCODE]}, ring {value_key} is missing')
        
    return pd.DataFrame(d)


def compute_mean(data, date_as_offset=False, ring_type='raw', biweight=False, id_mean=None, slope_resolution=0, min_elements=4):
    """
        sequences: DataFrame of samples. `offset` column is need if offset_type is `offset`
    """

    #print('mean', id_mean)
    if OFFSET not in data.columns:
        raise ValueError('mean: no offset in sequences')
    
    if date_as_offset:
        data[OFFSET] = data[DATE_BEGIN]

    if data[OFFSET].isna().any():
        key = DATE_BEGIN if date_as_offset else OFFSET
        raise ValueError(f'mean: one or more {key} contain NA values')
    
    if not (data[CATEGORY] == MEASURE).all():
        mask = data[CATEGORY] != MEASURE       
        if not data.loc[mask, MEANAS_MEASURE].all():
            raise ValueError('mean : data need to contain sample only')
    
    if not (data[DATA_TYPE] == ring_type).all():
        raise ValueError("mean: ring type don't match data")
        
    data_col = data2col(data)
    if biweight:
        median = data_col.median(skipna=True, axis=1)
        data_center = data_col.sub(median, axis=0)
        kind_std = data_center.abs().median(skipna=True, axis=1)
        data_center_reduce = data_center.div((biweight_factor * kind_std) + epsilon, axis=0)
        data_weights = (1 - (data_center_reduce ** 2)) ** 2
        data_weights[np.abs(data_center_reduce) >= 1] = 0
        weights = np.sum(data_weights, axis=1)
        means = np.sum(data_weights*data_col, axis=1) / weights
    else:
        means = data_col.mean(axis=1)
        weights = data_col.count(axis=1)
    
    sig = signatures(data_col, slope_resolution, min_elements)
    offsets = [(id, row[OFFSET]) for id, row in data.iterrows()]
    return np.round(means, 3).to_numpy(), weights.to_numpy(), offsets, sig, id_mean

def compute_means(data_dict, date_as_offset=False, ring_type='raw', biweight=False, num_threads=1, slope_resolution=0, min_elements=4):
    res = dict()
    #if num_threads == 1 or ProcessPoolExecutor is None:
    for id, data in data_dict.items():
        means, weights, offsets, signatures, id = compute_mean(data, date_as_offset=date_as_offset, ring_type=ring_type, biweight=biweight, id_mean=id, slope_resolution=slope_resolution, min_elements=min_elements)
        res[id] = (means, weights, offsets, signatures)
    # else:
    #     #perror(f'means multithreading: num_threads {num_threads}')
    #     with ProcessPoolExecutor(max_workers=num_threads) as executor:
    #         futures = []
    #         # Start an asynchronous task (future) for each index
    #         for id, data in data_dict.items():
    #             future = executor.submit(mean, data, date_as_offset=date_as_offset, ring_type=ring_type, biweight=biweight, id_mean=id)
    #             futures.append(future)

    #         # Wait for all tasks to be completed
    #         for future in futures:
    #             means, weights, offsets, id = future.result()
    #             res[id] = (means, weights, offsets)
    return res





"""
Application Logger
"""

__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"


import logging
import sys
import os
import panel as pn
import numpy as np
import webbrowser
import urllib.parse
from pathlib import Path
import json

GENERAL_LEVEL = logging.ERROR
APP_LEVEL = logging.DEBUG
NOTIFICATION_LEVEL = logging.INFO
LOG_PATH = Path(os.path.expanduser("~")) / 'pyDendron'
LOG_FILENAME = str(LOG_PATH / 'pydendron.log')
FORMAT = '%(asctime)s | %(levelname)s | %(name)s | %(message)s'

def get_git_version():
    import subprocess
    
    try:
        import importlib.metadata

        version = importlib.metadata.version('pyDendron')
        return version
    except ImportError:
        # Get the directory of the __init__.py file
        current_directory = Path(__file__).resolve().parent
        file_path = current_directory / 'version.txt'
        # Exécute la commande git pour obtenir la version (tag) la plus récente
        version = subprocess.check_output(['git', 'describe', '--tags']).strip().decode('utf-8')

        return version
    except subprocess.CalledProcessError:
            return "Number unable."
    
__version__ = get_git_version()


class NotificationStream():
    """
    A class that represents a notification stream.

    This class is responsible for writing log messages to the notification system.
    It provides methods to send different types of notifications based on the log message level.

    Args:
        duration (int, optional): The duration of the notifications in milliseconds. Defaults to 5000.
    """

    def __init__(self, duration=5000):
        self.duration = duration
        self.previous_message = None

    def _extracte_message(self, msg):
        """
        Extracts the relevant part of the log message.

        Args:
            msg (str): The log message.

        Returns:
            str: The extracted log message.
        """
        
        #if msg.find('[STDERR]') >= 0:
        #    return None
        #index = np.max(np.array([msg.find("DEBUG"), msg.find("WARNING"), msg.find("INFO"),
        #    msg.find("ERROR"), msg.find("CRITICAL")]))
        #if index > -1:
        #    msg = msg[index:]
        msg = msg.split('|')[-1]
        if len(msg) > 103:
            msg = msg[:50]+ '...'+msg[-50:]
        return msg
 
    def write(self, msg):
        """
        Writes a log message to the notification system.

        This method processes the log message and sends the appropriate notification based on the log message level.

        Args:
            msg (str): The log message to be sent as a notification.
        """
        if pn.config.notifications and (pn.state.notifications is not None):
            if msg is None:
                return
            text = self._extracte_message(msg)
            if text == self.previous_message:
                return
            if msg.find('DEBUG') >= 0:
                pn.state.notifications.send(text, background='hotpink', icon='<i class="fas fa-bug"></i>')
            elif msg.find('WARNING') >= 0:
                if msg.find('Dropping a patch') >= 0: #can't remove Bokeh message !!!
                    return
                pn.state.notifications.warning(text, self.duration*2)
            elif msg.find('INFO') >= 0:
                pn.state.notifications.info(text, self.duration)
            elif msg.find('ERROR') >= 0:
                pn.state.notifications.error(text, 0)
            elif msg.find('CRITICAL') >= 0:
                pn.state.notifications.send(text, duration=0, background='black', icon='<i class="fas fa-burn"></i>')
            else:
                #print('Unknown log message level:|||', msg, '|||')
                pn.state.notifications.send(text, duration=0, background='gray', icon='<i class="fas fa-bolt"></i>')
            self.previous_message = text

def add_stream(logger, handler, level=logging.INFO, format=FORMAT):
    handler.setLevel(level)
    handler.setFormatter(logging.Formatter(format)) # add formatter to ch
    logger.addHandler(handler) # add ch to logger
    return handler

class StderrToLogger(object):
    def __init__(self, logger, level=logging.DEBUG):
        self.logger = logger
        self.level = level

    def write(self, message):
        if (message.strip() != "") and (message.strip() != "^"):
            self.logger.log(self.level, '[STDERR] ' + message.rstrip())

    def flush(self):
        pass

def print_logger_info():
    print('*'*80)
    for name, logger in logging.Logger.manager.loggerDict.items():
        if isinstance(logger, logging.Logger):
            print('---->', name, 'cst general:', GENERAL_LEVEL, 'cst app:', APP_LEVEL, 'cur level:', logger.level, 'parent:',logger.parent, 'propagate:', logger.propagate)
            for handler in logger.handlers:
                print('\t---->', handler)
    print('*'*80)

def general_logger(level=GENERAL_LEVEL, format=FORMAT):
    logging.basicConfig(format=FORMAT, level=GENERAL_LEVEL, force=True)
    logging.captureWarnings(True)
    
    for name, logger in logging.Logger.manager.loggerDict.items():
        if isinstance(logger, logging.Logger):
            logger.setLevel(level)
            for handler in logger.handlers:
                handler.setLevel(level)
                handler.setFormatter(logging.Formatter(format))

    # create general logger
    logger_root = logging.getLogger()

    # create file handler
    LOG_PATH.mkdir(parents=True, exist_ok=True)
    file_handler = logging.FileHandler(LOG_FILENAME, mode='w')
    add_stream(logger_root, file_handler, level=level, format=format)

    return logger_root

def app_logger(level=APP_LEVEL, format=FORMAT):
    logger = logging.getLogger('pyDendron')
    logger.setLevel(level)

    # redirect stderr to logger
    sys.stderr = StderrToLogger(logger)

    # create notification stream to redirect log messages to the panel notification system
    notification_stream = NotificationStream()
    notification_handler = add_stream(logger_root, logging.StreamHandler(notification_stream), level=level, format=format)

    return logger, notification_stream

logger_root = general_logger()
logger, notification_stream = app_logger()

"""
-------------------------------------------------------------------------------
Logger addones functions
-------------------------------------------------------------------------------
"""
def catch_bokeh_log(cb_connection, cb_disconnection, app):
    """
    Catch the disconnection of a WebSocket connection and save the data.

    Args:
    
        cb_auto_save : callback to save data.
    """
    
    class BohehFilter(logging.Filter):
        def __init__(self, cb_connection, cb_disconnection, app):
            super(BohehFilter, self).__init__()
            self.cb_connection = cb_connection
            self.cb_disconnection = cb_disconnection
            self.app = app
            
        def filter(self, record):
            if record.getMessage().startswith('WebSocket connection closed'):
                self.cb_disconnection(self.app)
            elif record.getMessage().startswith('ServerConnection created'):
                self.cb_connection(self.app)
            return True

    logger_bokeh = logging.getLogger('bokeh.server.views.ws')
    logger_bokeh.setLevel(APP_LEVEL)
    logger_bokeh.addFilter(BohehFilter(cb_connection, cb_disconnection, app))

def catch_tornado_log(app):
    class TornadoFilter(logging.Filter):
        def __init__(self, app):
            super(TornadoFilter, self).__init__()
            self.app = app
            
        def filter(self, record):
            if record.getMessage().startswith('Exception in callback functools.partial(<bound method IOLoop._discard_future_result'):
                return False
            return True
    
    logger_tornado = logging.getLogger('tornado.application')
    logger_tornado.addFilter(TornadoFilter(app))

def open_mail_client(to, subject, body):
    mailto_link = f"mailto:{to}?subject={urllib.parse.quote(subject)}&body={urllib.parse.quote(body)}"
    webbrowser.open(mailto_link)

def perror(*args, **kwargs):
    """
    Print the message to sys.stderr with the same arguments as the print function.

    Parameters:
    *args : Variable length argument list to print.
    **kwargs : Arbitrary keyword arguments for print.
    """
    logger.debug(' '.join([repr(arg) for arg in args])+' '.join([f"{key}={repr(value)}" for key, value in kwargs.items()]))

    #print(*args, file=sys.stderr, **kwargs)
    
    
def check_version(cfg_filename):
    if Path(cfg_filename).is_file():
        with open(cfg_filename, 'r') as f:
            data = json.load(f)
            if 'version' in data:
                v = data["version"]
                if v == __version__:
                    return True
                else:
                    logger.info(f'ignore {cfg_filename}, versions differ: {v} != {__version__}')
    else:
        logger.info(f'CFG file not found: {cfg_filename}')
    
    return False



"""
    Import from DCCD dataverse
"""
__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"


from pyDataverse.api import NativeApi, DataAccessApi, SearchApi
from pyDataverse.models import Dataverse
import json
from pathlib import Path
import xml.etree.ElementTree as ET
import requests
from requests.exceptions import RequestException
import pyDendron.dataname as dn

path = Path('./dataset/dccd')
base_url = 'https://dataverse.nl'
search_api = SearchApi(base_url)
api_token = ''
api = NativeApi(base_url, api_token)
access_api = DataAccessApi(base_url, api_token)

def get_dccd(keywords='dccd'):
    dataverse_list = []
    suffix = keywords
    data_type = 'dataverse'
    start = 0
    total = 1
    while start < total:
        query = suffix + "&start=" + str(start)
        search_results = search_api.search(query, data_type=data_type).json()
        total = search_results['data']['total_count']
        start = search_results['data']['start']
        for i, result in enumerate(search_results['data']['items']):
            identifier = result['identifier']
            name = result['name']
            #print(i, total, identifier, name)
            dataverse_list.append({'type': 'dataverse', 'id': identifier, dn.KEYCODE:  name})
            start += 1
    return dataverse_list

def get_dataverse(identifier='dccd'):
    contents = api.get_dataverse_contents(identifier).json()
    return contents['data']

def get_dataset(pid, keycode, identifier):
    contents = api.get_dataset(pid).json()
    if 'latestVersion' in contents['data']:
        for i, content in enumerate(contents['data']['latestVersion']['files']):
            #print(i, content)
            if content['restricted'] == False:
                dir = path / Path(keycode) / Path(identifier)
                destination = dir / Path(content['dataFile']['filename']) 
                if not destination.exists():
                    req = access_api.get_datafile(content['dataFile']['id'])
                    dir.mkdir(parents=True, exist_ok=True)
                    #print(f'\t download {destination}')
                    req.raise_for_status()  
                    with open(destination, 'wb') as fic:
                        fic.write(req.content)
            else :
                logger.info(f'\t restricted {content["label"]}')
            
dccds = get_dccd()
dataverses = get_dataverse(dccds[0]['id'])
keycode = 'empty'
for dataverse in dataverses:
    keycode = dataverse[dn.KEYCODE]
    #print('-'*10)
    #print(f'access to dataset of {keycode}')
    datasets = get_dataverse(dataverse['id'])
    for dataset in datasets:
        if dataset['type'] == 'dataset':
            #print(f'get file of {dataset["identifier"]} from {keycode}')
            pid = f'{dataset["protocol"]}:{dataset["authority"]}/{dataset["identifier"]}'
            get_dataset(pid, keycode, dataset["identifier"])
        



""" I/O Heidelberg file format. """

__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"
__license__ = "GPL"

from pathlib import Path
import pandas as pd
import numpy as np
import re
import io

from pyDendron.dataname import *
from pyDendron.dataset import Dataset
from pyDendron.alien.io import IO
from pyDendron.app_logger import logger, perror

"""
From TRICYCLE: A UNIVERSAL CONVERSION TOOL FOR DIGITAL MEASURE-RING DATA – SUMMARY OF DENDRO DATA FORMATS
The header block begins with a line HEADER:. This is followed by lines of metadata, with one field on each line, in the format keywords=value much like a standard Windows INI file. As mentioned previously there are a number of predefined keywords, all of which are outlined here:

• KeyCode, KeyNo
• LabotaryCode
• Project, ProtectionCode

• BHD
• Bibliography, Bibliography[n], BibliographyCount
• Bundle
• MeanType, ChronoMemberCount, ChronoMemberKeycodes 

• Age
• Bark, Pith, JuvenileWood, SapWoodRings
• Circumference

• PersId
• Client, ClientNo
• Collector
• CreationDate
• Comment, • Comment[n], CommentCount

• Continent, Country, Town, TownZipCode, Street, District, State, Province
• CardinalPoint, Latitude, Longitude, Elevation, Exoffset

• CoreNo

• DataFormat, DataType

• AcceptDate
• DateBegin
• Dated
• DateEnd, DateEndRel, DateOfSampling
• DateRelBegin[n], DateRelEnd[n], DateRelReferenceKey[n], DateRelCount
• LastRevisionDate, LastRevisionPersID
• Location, LocationCharacteristics

• DeltaMissingRingsAfter, DeltaMissingRingsBefore, DeltaRingsFromSeedToPith 
• Disk
• 
• EdgeInformation
• EffectiveAutoCorrelation, EffectiveMean, EffectiveMeanSensitivity, EffectiveNORFAC, AutoCorrelation 
• StandardDeviation

• Key
• EffectiveNORFM, EffectiveStandardDeviation • Eigenvalue
• 
• EstimatedTimePeriod

• FieldNo
• FilmNo
• FirstMeasurementDate, FirstMeasurementPersID
• FromSeedToDateBegin
• GlobalMathComment[n], GlobalMathCommentCount 
• GraphParam
• Group
• HouseName, HouseNo
• ImageCellRow, ImageComment[n], ImageFile[n], ImageCount, ImageFile

• Interpretation
• InvalidRingsAfter • InvalidRingsBefore 
• 
• 
• LeaveLoss
• Length
• 
• MajorDimension
• MathComment, MathComment[n], MathCommentCount
• MeanSensitivity
• MinorDimension
• MissingRingsAfter, MissingRingsBefore

• NumberOfSamplesInChrono • NumberOfTreesInChrono

• 
• QualityCode
• Radius, RadiusNo
• RelGroundWaterLevel
• RingsFromSeedToPith
• SampleType, SamplingHeight, SamplingPoint
• Sequence
• SeriesEnd, SeriesStart, SeriesType
• ShapeOfSample
• Site, SiteCode
• SocialStand
• SoilType
• Species, SpeciesName
• StemDiskNo 
• Tree, TreeHeight
• Timber, TimberHeight, TimberType, TimberWidth
• TotalAutoCorrelation, TotalMean, TotalMeanSensitivity, TotalNORFAC, TotalNORFM, TotalStandardDeviation
• TreeNo
• Unit
• UnmeasuredInnerRings, UnmeasuredOuterRings 
• WaldKante
• WoodMaterialType
• WorkTraces

The header section is followed by a data section denoted by a line containing the keyword DATA: followed by the type of data present which can be one of Tree; HalfChrono; Chrono; Single; Double; Quad. Tree, HalfChrono and Chrono are the original keywords supported by early versions of TSAP but these are now deprecated in preferences of the more generic Single, Double and Quad terms. The terms Single, Double and Quad are largely interchangeable with Tree, HalfChrono and Chrono respectively, but not completely. Double can refer to both Tree and HalfChrono format data. When the newer terms are used, the header keyword DataFormat is used to record whether the data is equivalent to Tree, HalfChrono or Chrono.
Single format - data is typically used for storing raw measurement series. Each data line contains 10 data values each being a left space padded integer taking up 6 characters. Any spare data values in the final data line are filled with zeros. Alternatively it appears that TSAP-Win also accepts this data section as single integer values one per line.
Double format - data is for storing data with sample depth information - typically means. Like the single format section, data is stored as 10 integer values, each taking up 6 characters and left padded with spaces. The values are in pairs of ring-widths and sample depths, therefore five rings are stored per line.
Quad format - data is for storing means with sample depth as well as data on how many of the constituent series increase and decrease. This format therefore requires four numbers for each data point: ring-width; sample depth; increasing series; decreasing series. Numbers are stored as integers, left space padded as before, but this time only using 5 characters not 6. Four data points are included on each line, therefore this means there are 16 numbers per row and each row is 80 characters long.
"""

class IOHeidelberg(IO):
    file_extension = ['.fh']
    
    # Define regex patterns as class variables to compile them only once
    LONGITUDE_PATTERN = re.compile(r'(\d+). (\d+)\' (\d+)\'\' ([NS])')
    LATITUDE_PATTERN = re.compile(r'(\d+). (\d+)\' (\d+)\'\' ([WE])')

    def __init__(self, **params):
        super().__init__(**params)
        self.keys = self.init_keys()

    @staticmethod
    def _location(value):
        return value
    
    @staticmethod
    def _boolean(value):
        return value.upper() in ['TRUE', '+', '1', 'P', 'B']
    
    @staticmethod
    def _integer(value):
        return int(value) if value != '' else pd.NA

    @staticmethod
    def _date(value):
        return pd.to_datetime(value) if value != '' else pd.NA

    @staticmethod
    def _longitude(value):
        if not value:
            return pd.NA
        match = IOHeidelberg.LONGITUDE_PATTERN.match(value)
        if match:
            lon = match.group(1) + match.group(2) / 60 + match.group(3) / 3600        
            return -lon if match.group(4) == 'S' else lon
        return float(value)

    @staticmethod
    def _latitude(value):
        if value == '':
            return pd.NA
        match = IOHeidelberg.LATITUDE_PATTERN.match(value)
        if match:
            lat = match.group(1) + match.group(2) / 60 + match.group(3) / 3600        
            return -lat if match.group(4) == 'W' else lat
        return float(value)

    @staticmethod
    def _string(value):
        return value

    @staticmethod
    def _float(value):
        return float(value)
    
    @staticmethod
    def _list(value):
        return [x for x in re.split(r',|\s', value) if x != '']

    def init_keys(self):
        return {
        'Project': (PROJECT, self._string),
        'Keycode': (KEYCODE, self._string),
        'Bark': (BARK, self._boolean), 
        'Pith': (PITH, self._boolean),
        'Species': (SPECIES, self._string),
        #'MissingRingsAfter': (CAMBIUM_MAXIMUM, integer_), 
        #'MissingRingsBefore': (PITH_MAXIMUM, integer_),
        #'Location': (SITE_COUNTRY, location),
        'Country': (SITE_COUNTRY, self._string),
        'Town': (SITE_TOWN, self._string),
        'TownZipCode': (SITE_ZIP, self._string),
        'State': (SITE_STATE, self._string),
        'Province': (SITE_DISTRICT, self._string),
        'Latitude': (SITE_LATITUDE, self._latitude),
        'Longitude': (SITE_LONGITUDE, self._longitude),
        'Elevation': (SITE_ELEVATION, self._float),
        'CreationDate': (CREATION_DATE, self._date),
        'SapWoodRings': (SAPWOOD, self._integer),
        'DateEnd': (DATE_END, self._integer),
        'DateBegin': (DATE_BEGIN, self._integer),
        'Length': (DATA_LENGTH, self._integer),
        'ChronoMemberKeycodes': (None, self._list),
        }

    def read_header(self, buffer, meta):
        
        chrono_member_keycodes = None
        lines = buffer.split('\n')
        for line in lines:
            s = line.split('=')
            #print('read_header:', line, s)
            key = s[0] 
            value = s[1] if len(s) > 1 else ''
            if key in self.keys:
                (k, fct) = self.keys[key]
                if k is None:
                    chrono_member_keycodes = fct(value)
                else:
                    x = fct(value)
                    if pd.notna(x):
                        meta[k] = x
            elif key != '':
                logger.warning(f'key "{key}" with value "{value}" is not imported')
        
        if (DATE_BEGIN in meta ) and (meta[DATE_BEGIN] < 0): meta[DATE_BEGIN] += 1
        if (DATE_END in meta ) and (meta[DATE_END] > 0): meta[DATE_END] += 1
        if SAPWOOD in meta: meta[SAPWOOD] = meta[DATA_LENGTH] - meta[SAPWOOD]
        #if PITH_MAXIMUM in meta: meta[PITH_MAXIMUM] *= -1
        if (SITE_ELEVATION in meta) and (SITE_LONGITUDE in meta): self._get_location(meta)
        
        return chrono_member_keycodes
         
    def read_data(self, buffer, meta, kind):
        kind = kind.upper()
        dec = 1
        meta[CATEGORY] = MEASURE
        #print(meta)
        if kind in ['DOUBLE', 'HALFCHRONO']:
            kind = 2
            meta[CATEGORY] = MEAN
        elif kind in ['QUAD', 'CHRONO']:
            kind = 4
            meta[CATEGORY] = MEAN
        words = buffer.upper().split()   
        values = [] 
        for i, word in enumerate(words):
            if i * dec >= meta[DATA_LENGTH]:
                break
            if i % dec == 0:
                v = int(word)
                if v == 0:
                    values.append(pd.NA)
                else:
                    values.append(v)

        meta[DATA_VALUES] = np.array(values, dtype='float')
        meta[DATA_TYPE] = RAW
    
    def read_sequences(self, id_parent, lines):
        meta = {}
        state = 'start'
        buffer = ''
        id_comps = {}
        kind = 'single'
        chrono_member_keycodes = None
        
        for line in lines:
            #print(state, line)
            if line.upper().startswith('HEADER:'): # new sequences
                if state == 'data':
                    #print(meta)
                    self.read_data(buffer, meta, kind)
                if state != 'start':
                    self.components.append({ID_PARENT: id_parent, ID_CHILD: meta[ID], OFFSET: pd.NA})
                    self.sequences.append(meta)
                    id_comps[meta[KEYCODE]] = meta[ID]
                    if chrono_member_keycodes is not None:
                        for keycode in chrono_member_keycodes:
                            if keycode not in id_comps:
                                raise ValueError(f'"{keycode}" is missing')
                        for keycode in chrono_member_keycodes:
                            self.components.append({ID_PARENT: meta[ID], ID_CHILD: id_comps[keycode], OFFSET: pd.NA})

                meta = {ID: self.next_id()}
                state = 'header'                
                buffer = ''
            elif line.upper().startswith('DATA:'): # new values
                if state != 'header':
                    raise ValueError('inconsistent state: {state}')
                kind = line.split(':')[1]
                chrono_member_keycodes = self.read_header(buffer, meta)
                state = 'data'          
                meta[DATA_TYPE] = RAW
                buffer = ''
            else:
                buffer += line
                
        if state == 'header':
            raise ValueError('inconsistent state: {state}')
        elif state != 'data':
            self.read_data(buffer, meta, kind)
            self.components.append({ID_PARENT: id_parent, ID_CHILD: meta[ID], OFFSET:pd.NA})
            self.sequences.append(meta)

    def write_file(self, data: pd.DataFrame, means, filename: str):
        with open(Path(filename), mode='w', encoding=self.encoding, errors='ignore') as fd:
            for _, row in data.iterrows():
                if row[CATEGORY] in [MEASURE, MEAN]:
                    fd.write('HEADER:\n')
                    for key, (col, _) in self.keys.items():
                        if col in row:
                            v = row[col]
                            if pd.notna(v):
                                fd.write(f'{key}={v}\n')
                    if row[DATA_TYPE] != RAW:
                        fd.write(f'{DATA_TYPE}={row[DATA_TYPE]}\n')
                    fd.write(f'Lenght={row[DATA_LENGTH]}\n')
                    fd.write(f'DATA:Tree\n')
                    tmp = np.round(np.nan_to_num(row[DATA_VALUES], nan=0), 0).astype(int).tolist()
                    for i, v in enumerate(tmp):
                        fd.write(f'{v} ')
                        if (i + 1) % 10 == 0:
                            fd.write('\n')
                    while (i + 1) % 10 != 0:
                        fd.write('0 ')
                        i += 1
                    fd.write('\n')
                            
                        
                    



"""
ComponentsTree class
"""
__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"
__license__ = "GPL"




import logging

import pandas as pd
import numpy as np
import param

from collections import Counter
from pyDendron.app_logger import logger, perror
from pyDendron.dataname import *

class ComponentsNode():
    def __init__(self, parent, id, keycode, category, offset, depth=0):
        self.id = id
        self.keycode = keycode
        self.parent = parent
        self.category = category
        self.offset = offset
        self.depth = depth
        self.descendants = {MEASURE:{}, MEAN:{}, SET:{}}
        if category in CATEGORIES:
            self.descendants[category][self] = self.offset
        self.children = []

    def append(self, node):
        self.children.append(node)
        for category in node.descendants:
            for child, offset in node.descendants[category].items():
                self.descendants[category][child] = self.offset + offset
        
    def get_children(self, current=False, max_depth=None):
        if max_depth is not None: 
            d = {node: node.offset+self.offset for node in self.children if node.depth <= max_depth}
        d = {node: node.offset+self.offset for node in self.children}
        if current:
            d[self] = self.offset
        return d

    def filter(self, categories=None, max_depth=pd.NA):
        def _descendants(category, depth):
            if pd.notna(max_depth): 
                return {node: offset for node, offset in self.descendants[category].items() if (node.depth-self.depth) <= depth}
            return self.descendants[category]
        
        d = {}
        if categories is None:
            categories = CATEGORIES
        for category in categories:
            d.update(_descendants(category, max_depth))
        return d

    def count_descendants(self, category=None):
        def count_category(category):
            lst = [node.id for node in self.descendants[category].keys()]
            keycodes = {node.id: node.keycode for node in self.descendants[category].keys()}
            cpt = Counter(lst)
            return cpt
        
        if category is not None:
            return count_category(category)
        else:
            l = {}
            for category in CATEGORIES:
                l.update(count_category(category))
            return l


    def detect_duplicates(self, category=None, raise_error=True):
        def duplicates(category, raise_error):
            lst = [node.id for node in self.descendants[category].keys()]
            keycodes = {node.id: node.keycode for node in self.descendants[category].keys()}
            if len(lst) != len(set(lst)):
                cpt = Counter(lst)
                dup = [(element, keycodes[element]) for element, count in cpt.items() if count > 1]
                if raise_error:
                    raise ValueError(f'detect_duplicates: duplicate samples : {dup}')
                return dup
            return []
        
        if category is not None:
            return duplicates(category, raise_error)
        else:
            l = []
            for category in CATEGORIES:
                l += duplicates(category, raise_error)
            return l

    def find(self, ids, depth=pd.NA):
        l = []
        sids = set(ids)
        for node in self.children + [self]:
            if node.id in sids:
                l.append(node.id)
        return l

    def print(self, descendants=False):
        print('\t'*self.depth, f'id: {self.id}, category: {self.category}, offset: {self.offset}, depth: {self.depth}') 
        if descendants : 
            print('\t'*self.depth,self.descendants)
        for n in self.children:
            n.print()

class ComponentsTree(ComponentsNode):
    def __init__(self):
        super(ComponentsTree, self).__init__(None, -1, 'Root', 'TreeNode', 0)   

    def stats(self):
        return pd.DataFrame([{
        'roots' : len(self.get_mean_ids(0)),
        'chonologies' : len(self.get_mean_ids()),
        'set' : len(self.get_set_ids()),
        'samples' : len(self.get_samples()),
        'duplicated chonologies' : len(self.detect_duplicated_means(raise_error=False)),
        'duplicated samples' : len(self.detect_duplicated_samples(raise_error=False))
        }])

    


"""
Ploter panel
"""
__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"

from pathlib import Path
import json

import panel as pn
from panel.viewable import Viewer
from bokeh.models import ColumnDataSource

from pyDendron.app_logger import logger, perror, __version__, check_version
from pyDendron.dataname import *
from pyDendron.dataset import Dataset
from pyDendron.gui_panel.dataset_package import DatasetPackage

from pyDendron.ploter import Ploter


class PloterPanel(Viewer):

    def __init__(self, dataset, parameters, cfg_path, **params):
        super(PloterPanel, self).__init__(**params)   
        bt_size = 75
        self.dataset = dataset
        self.parameters = parameters
        self.cfg_path = cfg_path
            
        self.dataset_package = DatasetPackage(dataset, parameters, name='ploter', editable=True, orderable=True)
        self.dataset_package.param.watch(self._sync_data, ['notify_package_change'], onlychanged=True)

        self.ploter = self.load_cfg()
        
        self.sliding_select = pn.widgets.Select(name='Sliding serie', options=[], description='Select the serie to slide.')
        self.sliding_select.param.watch(self.on_sliding_select, ['value'], onlychanged=True)
        #self.ploter.param.watch(self.on_sliding_select, ['y_delta', 'y_height', 'x_offset_mode', 'anatomy', 
        #                                                'y_offset_mode', 'draw_type', 'cambium_estimation', 
        #                                                'legend', 'legend_location', 'show_dates', 'curve_axe',
        #                                                'show_estimated_dates', 'show_cambium_season',
        #                                                              ], onlychanged=True)
        
        self.ploter.param.watch(self._sync_data, ['data_type'], onlychanged=True)

        self.x_slider = pn.widgets.IntSlider(name='X Slider', start=0, end=8, step=1, value=0,
                                             sizing_mode='stretch_width', max_width=250)
        self.y_slider = pn.widgets.FloatSlider(name='Y Slider', start=0, end=8, step=1, value=0, 
                                             sizing_mode='stretch_width', max_width=250)
        
        self.x_slider.param.watch(self.on_slider_x, ['value'], onlychanged=True)
        self.y_slider.param.watch(self.on_slider_y, ['value'], onlychanged=True)
        self.bt_save = pn.widgets.Button(name='Save offset', icon='Save', button_type='primary', align=('start', 'center'), width=2*bt_size, description='Save the offset.' )
        self.bt_save.on_click(self.on_save_offset)
        
        self.bt_plot = pn.widgets.Button(name='Plot', icon='pencil', button_type='primary', align=('start', 'center'), width=2*bt_size, description='Replot the package.' )
        self.bt_plot.on_click(self.on_plot)
        
        layout_light = pn.Row(
            self.bt_plot,
            pn.Row('### Sliding',
                self.sliding_select,
                self.x_slider,
                self.y_slider,
                self.bt_save, styles=dict(background='WhiteSmoke')
            ))
        self._layout = pn.Column(self.dataset_package, layout_light, self.ploter, name=self.name,
                                  margin=(5, 0), sizing_mode='stretch_width')
    
    def on_save_offset(self, event):
        pass
    
    def get_sidebar(self, visible=True):
        pploter = pn.Param(self.ploter, name='Dynamic')
        #reploter = pn.Param(self.ploter.param_replot, name='Reset')
        return pn.Card(pploter, title='Plot', sizing_mode='stretch_width', margin=(5, 0), collapsed=True, visible=visible)  

    def dump_cfg(self):
            """
            Dump the configuration of the ploter panel to a JSON file.

            This method serializes the parameters of the ploter panel and saves them
            to a JSON file specified by `self.cfg_file`.

            Returns:
                None
            """
            cfg_file = self.cfg_path / Path(f'{self.__class__.__name__}.cfg.json')
            with open(cfg_file, 'w') as fd:
                data = {
                    'version': __version__,
                    'ploter' : self.ploter.param.serialize_parameters(),
                }
                json.dump(data, fd)

    def reset(self):
        self.ploter.clear()

    def load_cfg(self):
        """
        Load the configuration for the ploter.

        Returns:
            Ploter: The ploter object with the loaded configuration.
        """
        try:
            cfg_file = self.cfg_path / Path(f'{self.__class__.__name__}.cfg.json')

            if check_version(cfg_file):
                with open(cfg_file, 'r') as fd:
                    data = json.load(fd)
                    ploter = Ploter(**Ploter.param.deserialize_parameters(data['ploter']))
            else:
                ploter = Ploter()
        except Exception as inst:
            logger.warning(f'ignore {cfg_file } ploter panel. {inst}')
        return ploter

    def __panel__(self):
        return self._layout

    def _sync_data(self, event):
        try:
            data = self.dataset_package.data
            if data is not None:
                #logger.debug(f'ploter panel: sync data --> data not None')
                lst = self.dataset_package.data[KEYCODE].to_list() 
                self.sliding_select.options = ['None'] + lst
                #self.on_replot(None)
            else:
                #logger.debug(f'ploter panel: sync data --> data None')
                self.sliding_select.options = ['None']
            self.reset()
            self.sliding_select.value = 'None'
        except Exception as inst:
            logger.error(f'ploter panel: {inst}', exc_info=True)

    def on_plot(self, event):
        self._layout.loading = True
        #self.ploter.visible = False
        try:
            #perror('on_plot before', self.ploter.width, self.ploter.height, self.ploter.figure_pane.object.width, self.ploter.figure_pane.object.height)
            #self.ploter.clear()
            #logger.debug(f'ploter panel: on_replot')
            if self.ploter.data_type == 'Raw':
                self.ploter.prepare_and_plot(data=self.dataset_package.data)
            elif self.ploter.data_type == 'Detrend':
                self.ploter.prepare_and_plot(data=self.dataset_package.dt_data)
            elif self.ploter.data_type == 'Log':
                self.ploter.prepare_and_plot(data=self.dataset_package.log_data)
        except Exception as inst:
            logger.error(f'ploter panel: {inst}', exc_info=True)
        finally:
            #self.ploter.visible = True
            self._layout.loading = False
            #perror('on_plot after', self.ploter.width, self.ploter.height, self.ploter.figure_pane.object.width, self.ploter.figure_pane.object.height)


    def on_sliding_select(self, event):
        """
        Event handler for the sliding select widget.

        Parameters:
        - event: The event object triggered by the sliding select widget.

        Description:
        - This method is called when the value of the sliding select widget changes.
        - It updates the x and y sliders based on the selected value and the draw data.
        """
        if (self.sliding_select.value != 'None') and (self.ploter.draw_data is not None):
            keycode = self.sliding_select.value
            data = self.ploter.draw_data[keycode]
            fig = self.ploter.figure_pane.object
            self.x_slider.start =  int(fig.x_range.start - data[DATA_LENGTH])
            self.x_slider.end = int(data[DATA_LENGTH] + fig.x_range.end)
            self.x_slider.step = 1
            self.x_slider.value = data['x_offset']
            
            self.y_slider.start = fig.y_range.start - data['w_min'] 
            self.y_slider.end = fig.y_range.end + data['w_max']
            self.y_slider.value = data['y_offset']
            self.y_slider.step = (self.y_slider.end - self.y_slider.start) // 100
            
    def on_slider_x(self, event):
        """
        Event handler for the x-slider.

        Adjusts the x-offset of the plotted data based on the slider value.
        Updates the 'x' value kinds of the ColumnDataSource objects accordingly.

        Args:
            event: The event object triggered by the slider.

        Returns:
            None
        """
        if (self.ploter.draw_data is not None) and (self.sliding_select.value in self.ploter.draw_data):
            data_slide = self.ploter.draw_data[self.sliding_select.value]
            delta_x = data_slide['x_offset'] - self.x_slider.value
            data_slide['x_offset'] = self.x_slider.value
            for key, info in data_slide.items():
                if isinstance(info, ColumnDataSource):
                    for key in ['x', 'left', 'right', 'x0', 'x1']:
                        if key in info.data:
                            info.data[key] = [x - delta_x for x in info.data[key]]
    
    def on_slider_y(self, event):
        """
        Event handler for the y-slider.
    
        Adjusts the y-offset of the plotted data based on the slider value.
        Updates the 'y' value kinds of the ColumnDataSource objects accordingly.
    
        Args:
            event: The event object triggered by the slider.
    
        Returns:
            None
        """
        if (self.ploter.draw_data is not None) and (self.sliding_select.value in self.ploter.draw_data):
            fig = self.ploter.figure_pane.object
            info = self.ploter.draw_data[self.sliding_select.value]
            delta_y = info['y_offset'] - self.y_slider.value
            info['y_offset'] = self.y_slider.value

            for key, value in info.items():
                if isinstance(value, ColumnDataSource):
                    for key in ['y', 'top', 'bottom', 'y0', 'y1']:
                        if key in value.data:
                            value.data[key] = [y - delta_y for y in value.data[key]]
            
            old_y_mean = info['y_label']
            info['y_label'] = round(info['w_mean'] + info['y_offset'], 3)
            self.ploter.on_legend()




"""
Package builder
"""
__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"

import pandas as pd
import param
import panel as pn
import warnings
from panel.viewable import Viewer

from pyDendron.app_logger import logger, perror

from pyDendron.dataname import *
from pyDendron.gui_panel.dataset_package import save_package

from pyDendron.gui_panel.tabulator import (_cell_text_align, _cell_editors, _header_filters, _cell_formatters, 
                                           _get_selection)

class DatasetPackageBuilder(Viewer):
    selection = param.List(default=[], doc='path')

    def __init__(self, dataset, parameters, treeview, **params):
        bt_size = 80
        super(DatasetPackageBuilder, self).__init__(**params)   

        self.dataset = dataset
        self.parameters = parameters
        self.param_columns = self.parameters.columns
        self.treeview = treeview

        self.param_columns.get_widgets().param.watch(self.sync_columns, ['value'], onlychanged=True)

        self.wpath_local = pn.Row()
        self.wpath_treeview = self.treeview.wpath
        
        self.wpath_treeview.param.watch(self.sync_path, ['objects'], onlychanged=True)
        
        self.sql_field = pn.widgets.AutocompleteInput(name='Column', min_characters=0, options=self.param_columns.get_columns_hiddens(), placeholder='Select a column', width=bt_size*2, description='Select a column to filter.')
        self.sql_field.param.watch(self.sync_column_value, ['value'], onlychanged=True)
        self.sql_operator = pn.widgets.Select(name='operator', options=['==', '>', '<', '!=', 'contains' ], width=bt_size*2, description='Select an operator for filter.')
        self.sql_value = pn.widgets.AutocompleteInput(name='value', min_characters=0, restrict=False, options=[], placeholder='Enter a value', width=bt_size*2, description='Enter a value for filter.')
        self.bt_and_add = pn.widgets.Button(name='And ...', icon='logic-and', button_type='primary', width=int(1.5*bt_size), align=('start', 'end'), description='Add the filter with an AND condition.')
        self.bt_and_add.on_click(self.on_add_and)
        self.bt_or_add = pn.widgets.Button(name='Or ...', icon='logic-or', button_type='primary', width=int(1.5*bt_size), align=('start', 'end'), description='Add the filter with an OR condition.')
        self.bt_or_add.on_click(self.on_add_or)

        self.wfilter = pn.widgets.TextAreaInput(name='Filter', value=f'(`{CATEGORY}` != "{SET}")', height=60, sizing_mode='stretch_width', description='SQL like expression to apply to the rows selection.')

        self.bt_select = pn.widgets.Button(name='Append', icon='table-plus', button_type='primary', width=int(1.75*bt_size), align=('start', 'center'))
        self.bt_select.on_click(self.on_select)
        self.rb_append_mode = pn.widgets.RadioBoxGroup(name='Append mode', options=['empty table', 'current table'], inline=True, align=('start', 'center'))
        self.tg_level = pn.widgets.RadioBoxGroup(name='Level', value='current level', options=['current level', 'all descendants'], inline=True, align=('start', 'center'))
        
        self.add_parent = pn.widgets.Checkbox(name='Add parent', value=False, align=('start', 'center'))
        
        dtype = self.param_columns.get_dtype()
        self.wtabulator = pn.widgets.Tabulator(pd.DataFrame(columns=self.param_columns.get_columns_hiddens()), name='Result',                 
                hidden_columns=self.param_columns.get_hiddens(), 
                pagination='local',
                selectable='checkbox', 
                sizing_mode='stretch_width',
                text_align=_cell_text_align(dtype),
                editors=_cell_editors(dtype), 
                frozen_columns=[ICON, KEYCODE], 
                header_filters=_header_filters(dtype), 
                formatters=_cell_formatters(dtype),
                min_height=300,
                page_size=1000,
                max_height=300,
                height_policy='max',
                )
        self.bt_erase = pn.widgets.Button(name='Remove row', icon='eraser', button_type='primary', width=int(1.75*bt_size), align=('start', 'end'), description='Remove the selected rows.')
        self.bt_erase.on_click(self.on_erase)
        
        self.bt_save = pn.widgets.Button(name='Save package', icon='file', button_type='primary', width=int(1.75*bt_size), align=('start', 'end'), description='Save the package into the dataset.')
        self.bt_save.on_click(self.on_save)
        self.wselection_name = pn.widgets.TextInput(name='Name', sizing_mode='stretch_width', description='Package name.')
        
        
        self._layout = pn.Column(
                        self.wpath_local,
                        pn.Row(self.sql_field, self.sql_operator, self.sql_value, self.bt_and_add, self.bt_or_add), 
                                self.wfilter, #title='Select', collapsed=True, sizing_mode='stretch_width'),
                        pn.Row(self.bt_select, 
                               pn.pane.HTML('<b>In:</b>', align=('start', 'center')), pn.widgets.TooltipIcon(value="Clear table before append or keep current rows.", align=('start', 'center')), self.rb_append_mode, 
                               pn.pane.HTML('<b>From:</b>', align=('start', 'center')), pn.widgets.TooltipIcon(value="Select rows form the current path or from the root (all dataset).", align=('start', 'center')),self.tg_level,
                               pn.pane.HTML('<b>Option:</b>', align=('start', 'center')), pn.widgets.TooltipIcon(value="Include parent.", align=('start', 'center')), self.add_parent),
                                
                        self.wtabulator, 
                        self.bt_erase,
                        pn.Row(self.bt_save, self.wselection_name),
                        )

    def __panel__(self):
        return self._layout
    
    def sync_path(self, event):
        """
        Synchronizes the path in the GUI with the current dataset path.

        Parameters:
            event (Event): The event that triggered the synchronization.

        Returns:
            None
        """
        self.wpath_local.objects = [pn.pane.Markdown('**From current Dataset path:**', align=('start', 'center'), margin=(0, 0))] + self.wpath_treeview.objects[1:]
        
    def sync_columns(self, event):
        """
        Synchronizes the hidden columns in the tabulator widget based on the selected columns in the columns widget.

        Parameters:
        - event: The event object triggered by the user action.

        Returns:
        None
        """
        self.param_columns.update_tabulator(self.wtabulator)

    def build(self):
        """
        Builds the dataset package.

        Returns:
            tuple: A tuple containing the wtabulator and _layout objects.
        """
        return self.wtabulator, self._layout

    def sync_column_value(self, event):
        """
        Synchronizes the value of the 'wvalue' attribute based on the selected event.

        Args:
            event: The event object containing the new value.

        Returns:
            None
        """
        if event.new != '':
            try:
                self._layout.loading = True
                if event.new not in [DATA_INFO, DATA_VALUES, DATA_WEIGHTS, DATA_SIGNATURES]:
                    if self.param_columns.get_dtype()[event.new] != 'string':
                        self.sql_value.options = []
                    else:
                        self.sql_value.options = self.dataset.sequences[event.new].dropna().unique().tolist()
            except Exception as inst:
                logger.error(f'sync_column_value: {inst}', exc_info=True)
                self.sql_value.options = []
            finally:
                self._layout.loading = False

    def on_save(self, event):
        """
        Event handler for the save button.

        Saves the package using the values from the `wtabulator` and `wselection_name` widgets.

        Parameters:
        - event: The event object triggered by the save button.

        Returns:
        None
        """
        save_package(self.wtabulator.value, self.wselection_name.value, self.dataset)
            
    def on_add_filter(self):
        """
        Adds a filter to the dataset package builder.

        Returns:
            None
        """
        if self.param_columns.get_dtype()[self.sql_field.value] == 'string':
            if self.sql_operator.value == 'contains':
                self.wfilter.value += f'(`{self.sql_field.value}`.str.contains("{self.sql_value.value}"))'
            else:
                self.wfilter.value += f'(`{self.sql_field.value}` {self.sql_operator.value} "{self.sql_value.value}")'
        else:
            if self.sql_operator.value != 'contains':
                self.wfilter.value += f'(`{self.sql_field.value}` {self.sql_operator.value} {self.sql_value.value})'
        
    def on_add_and(self, event):
        """
        Appends ' and ' to select string.
        """
        self.wfilter.value += ' and '
        self.on_add_filter()
        
    def on_add_or(self, event):
        """
        Appends ' or ' to select string.

        Args:
            event: The event object triggered by the user action.
        """
        self.wfilter.value += ' or '
        self.on_add_filter()
    
    def on_erase(self, event):
        """
        Event handler for the erase button.

         It removes the selected rows from the wtabulator widget.

        Args:
            event: The event object triggered by the erase button click.

        Returns:
            None
        """
        try:
            self._layout.loading = True
            df_selection = _get_selection(self.wtabulator)
            self.wtabulator.value = self.wtabulator.value.drop(df_selection.index.to_list())
        except Exception as inst:
            logger.error(f'on_erase: {inst}', exc_info=True)
            self.wtabulator.value = pd.DataFrame(columns=list(self.param_columns.get_dtype().keys()))
        finally:
            self.wtabulator.selection = []
            self._layout.loading = False

    # def get_descendants(self, groups, groups_all, id_parent):
    #     """
    #     Recursively retrieves the descendants of a given parent index.

    #     Parameters:
    #         groups (pandas.core.groupby.DataFrameGroupBy): The groups object containing the hierarchical data.
    #         groups_all (pandas.core.groupby.DataFrameGroupBy): The groups_all object containing all the groups.
    #         id_parent (int): The index of the parent node.

    #     Returns:
    #         list: A list of indices representing the descendants of the parent node.
    #     """
    #     lst = []
    #     if id_parent in groups.groups:
    #         for id_parent, id_child in groups.get_group(id_parent).index:
    #             lst += self.get_descendants(groups, groups_all, id_child)
        
    #     if id_parent in groups_all.groups:
    #         lst += groups_all.get_group(id_parent).index.to_list()

    #     return lst

    def on_select(self, event):
        """
        Handle the event when an item is selected in the GUI treeview.

        Args:
            event: The event object representing the selection event.

        Returns:
            None

        Raises:
            None
        """
        try:
            self._layout.loading = True
            res = None
            
            if len(self.treeview.path) < 1:
                logger.warning('Creation of package not available if the path is empty (root).')
            else:
                id_parent = self.treeview.path[-1]
                id_roots= self.treeview.path[-2] if len(self.treeview.path) > 1 else None
                if self.tg_level.value == 'current level':
                    if self.add_parent.value:
                        if id_roots is None:
                            logger.warning('Add parent not possible with empty root')
                        res = self.dataset.get_data(id_parent, id_roots=id_roots, max_depth=1)
                    else:
                        res = self.treeview.wtabulator.value.set_index([ID_PARENT, ID_CHILD])
                    if self.wfilter.value != '':
                        res = res.query(self.wfilter.value)
                    self.wselection_name.value = self.wpath_local.objects[-1].name
                else:
                    if self.add_parent.value:
                        if id_roots is None:
                            logger.warning('Add parent not possible with empty root')
                        res = self.dataset.get_data(id_parent, id_roots=id_roots)
                    else:
                        res = self.dataset.get_data(id_parent)

                    if self.wfilter.value != '':
                        res = res.query(self.wfilter.value)
                    if self.rb_append_mode.value ==  'empty table':
                        self.wselection_name.value = self.wpath_local.objects[-1].name
                    else:
                        self.wselection_name.value = f'{self.wselection_name.value} + {self.wpath_local.objects[-1].name}'
                if res is not None:
                    if self.rb_append_mode.value ==  'empty table':
                        self.wtabulator.value = self.treeview.data.loc[res.index.to_list(), :].reset_index()
                    else:
                        with warnings.catch_warnings():
                            # TODO: pandas 2.1.0 has a FutureWarning for concatenating DataFrames with Null entries
                            warnings.filterwarnings("ignore", category=FutureWarning)
                            self.wtabulator.value = pd.concat([self.wtabulator.value, self.treeview.data.loc[res.index.to_list(), :].reset_index()])
                else:
                    logger.warning('on_select: empty result')
        except Exception as inst:
            logger.error(f'on_select: {inst}', exc_info=True)
            self.wtabulator.value = pd.DataFrame(columns=self.param_columns.get_columns_hiddens())
            
        finally:
            self._layout.loading = False



"""
Data Name
"""
__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"





import pandas as pd
import numpy as np
import copy


#--------------------------
# Dateset

## Categories
MEAN = 'Mean'
MEASURE = 'Measure'
SET = 'Set'
CATEGORIES = [MEAN, MEASURE, SET]

## Root in dataset
ROOT = -10
WORKSHOP = -20
CLIPBOARD = -30
TRASH = -40

## Dateset output keys (json, excel)
SEQUENCES = 'sequences'
COMPONENTS = 'components'
PACKAGES = 'packages'
VERSION = 'version'
LOG = 'log'
CROSSDATING = 'crossdating'

## Crossdating fileds
CROSSDATING_DATE = 'Crossdating Date'

## dataset check method
OFFSET_NORM = 'OffsetNorm'
DATE_BEGIN_NORM = 'BateBeginNorm'

## Components
### Components fileds
ID_PARENT = 'IdParent'
ID_CHILD = 'IdChild'
OFFSET = 'Offset'

### Components index
components_index = [ID_PARENT, ID_CHILD]

### Components type dictonary
components_dtype_dict = {
    OFFSET: 'Int32'
}

### Components columns list
components_cols = list(components_dtype_dict.keys())

### Components type list
components_dtype = list(components_dtype_dict.values())

## Sequences 
### Sequence fileds
ID = 'Id'
LABORATORY_CODE = 'LaboratoryCode'
PERS_ID = 'PersId'
PROJECT = 'Project'
KEYCODE = 'Keycode'
KEYCODE_MASTER = 'KeycodeMaster'
KEYCODE_PARENT = 'KeycodeParent'
SPECIES = 'Species'
SITE_LATITUDE = 'Latitude'
SITE_LONGITUDE = 'Longitude'
SITE_ELEVATION = 'Elevation'
SITE_CODE = 'SiteCode'
BIBLIOGRAPHY_CODE = 'BibliographyCode'
DATE_BEGIN = 'DateBegin'
#DATE_BEGIN_MASTER = 'DateBeginMaster'
DATE_END = 'DateEnd'
#SYNC = 'Sync'
DATE_BEGIN_ESTIMATED = 'DateBegin'
DATE_END_ESTIMATED = 'DateEnd'
CREATION_DATE = 'CreationDate'
DATE_SAMPLING = 'DateOfSampling'
CATEGORY = 'Category'
SUBCATEGORY = 'Subcategory'
MEANAS_MEASURE = 'MeanAsMeasure'
SAPWOOD = 'Sapwood' # not in data
PITH = 'Pith'
CAMBIUM = 'Cambium'
CAMBIUM_SEASON = 'CambiumSeason'
BARK = 'Bark'
COMMENTS = 'Comments'
URI = 'URI'
DATA_LENGTH = 'DataLength'
DATA_TYPE = 'DataType' 
DATA_VALUES = 'DataValues'
DATA_WEIGHTS = 'DataWeights'
DATA_INFO = 'DataInfo'
DATA_SIGNATURES = 'DataSignatures'
INCONSISTENT = 'Inconsistent'
TAG = 'Tag'
COMPONENT_COUNT = 'ComponentCount'

### Sequence index
sequences_index = [ID]

### Sequence type dictonary
sequences_dtype_dict = {
    KEYCODE: 'string', PROJECT: 'string', 
    SPECIES: 'string', 
    CATEGORY: 'string', SUBCATEGORY: 'string', 
    MEANAS_MEASURE : 'boolean',
    DATE_BEGIN: 'Int32', DATE_END: 'Int32', 
    SITE_LATITUDE: 'Float32', SITE_LONGITUDE: 'Float32', SITE_ELEVATION: 'Float32',
    SITE_CODE: 'string',
    PITH: 'boolean', 
    SAPWOOD: 'Int32', 
    CAMBIUM: 'boolean', 
    CAMBIUM_SEASON: 'string', 
    BARK: 'boolean', 
    CREATION_DATE: 'datetime64[ns]',
    DATE_SAMPLING : 'datetime64[ns]',
    LABORATORY_CODE : 'string',
    PERS_ID : 'string',
    BIBLIOGRAPHY_CODE : 'string',
    COMMENTS: 'string',
    URI: 'string',
    DATA_LENGTH : 'Int32', DATA_TYPE : 'string', DATA_VALUES : 'object',
    DATA_WEIGHTS : 'object', DATA_INFO : 'object', DATA_SIGNATURES : 'object',
    INCONSISTENT : 'boolean',
    TAG: 'string',
    COMPONENT_COUNT: 'Int32',
    
}

### Sequence columns list
sequences_cols = list(sequences_dtype_dict.keys())
### Sequence type list
sequences_dtype = list(sequences_dtype_dict.values())

## dataset log
### dataset log fileds
COLUMN = 'ColumnName'
DATE = 'UpdateDate'
NEWVALUE = 'NewValue'
OLDVALUE = 'OldValue'
USERNAME = 'UserName'

### dataset log dtype
log_dtype = {
    DATE: 'datetime64[ns]',
    ID_CHILD: 'Int32',
    ID_PARENT: 'Int32',
    COLUMN : 'String',
    OLDVALUE : 'object',
    NEWVALUE : 'object',
    USERNAME : 'String',
    COMMENTS: 'string',
}

#--------------------------
# Detrend

## Detrend methods
HANNING = 'Hanning'
HAMMING = 'Hamming'
BARTLETT = 'Bartlett'
BLACKMAN = 'Blackman'
RECTANGULAR = 'Rectangular'
BESANCON = 'Besancon (classic)'
BESANCON1 = 'besancon (log at the end)'
BP73 = 'BP73'
SPLINE = 'Spline'
SLOPE = 'Slope'
RAW = 'Raw'
CORRIDOR = 'Corridor (polynome)'
CORRIDOR_SPLINE = 'Corridor (spline)'
DELTA = 'Delta (first derivative)'
DELTADELTA = 'DeltaDelta (second derivative)'
LOG = 'log'

## detrend list
detrend_types = [RAW, HANNING, HAMMING, BARTLETT, BLACKMAN, RECTANGULAR, BESANCON, BESANCON1, 
             BP73, SPLINE, SLOPE, CORRIDOR, DELTA, DELTADELTA, LOG]

## detrend parameter fileds
DETREND = 'Detrend'
DETREND_WSIZE = 'Detrend Window Size'
DETREND_LOG = 'Detrend Log'

#--------------------------
# Mean

## Mean parameter fileds
MEANDATE_AS_OFFSET = 'Mean Date as Offset'
BIWEIGHT_MEAN = 'Biweight Mean'

#--------------------------
# CrossDating

## CrossDating fileds

ID_MASTER = 'IdMaster'
CORR_OVERLAP = 'r overlap'
GLK_OVERLAP = 'glk overlap'
DIST_OVERLAP = 'd overlap'

CORRELATION = 'correlation'
SYNC = 'Sync'

CORR = 'r'
GLK = 'glk'
#DIST = 'distance' 

T_SCORE = 't-score' 
Z_SCORE = 'z-score'

T_RANK= 't-rank'
Z_RANK= 'z-rank'
D_RANK= 'd-rank'

ZP_VALUE = 'zp-value'
TP_VALUE = 'tp-value'

CORR_OVERLAP_NAN = 'r Nnan'
GLK_OVERLAP_NAN = 'glk Nnan'
DIST_OVERLAP_NAN = 'd Nnan'

#PV = 'PV'
DCG = 'DCG'
AGC = 'agc'
SSGC = 'ssgc'
SGC = 'sgc'

COSINE = 'cosine'
EUCLIDEAN = '-1 x euclidean'
CITYBLOCK = '-1 x cityblock'
DISTANCE = 'distance'

crossdating_method = [CORRELATION, GLK, DISTANCE]
crossdating_distance = [COSINE, CITYBLOCK, EUCLIDEAN]

#--------------------------
# Location
SITE_COUNTRY = 'SiteCountry'
SITE_STATE = 'SiteState'
SITE_DISTRICT = 'SiteDistrict'
SITE_TOWN = 'SiteTown'
SITE_ZIP = 'SiteZipcode'

#--------------------------
# Drawing
HEARTWOOD = 'Heartwood'
MISSING_RING_BEGIN = 'MissingRingBegin'
MISSING_RING_END = 'MissingRingEnd'
CAMBIUM_BOUNDARIES = 'CambiumBoundaries'

CAMBIUM_ESTIMATED = 'CambiumEstimated'
CAMBIUM_LOWER = 'CambiumLower'
CAMBIUM_UPPER = 'CambiumUpper'

#PITH_ABSENT = 'PithAbsent'
PITH_ESTIMATED = 'PithEstimated'
PITH_LOWER = 'PithLower'
PITH_UPPER = 'PithUpper'

#--------------------------
# Treeview Tabulator

# Treeview Tabulator specific columns
ICON = 'Icon'
DATE_END_CE = '[B]CEEnd'
DATE_BEGIN_CE = '[B]CEBegin'

# Treeview Tabulator column types
dtype_view = sequences_dtype_dict.copy()
dtype_view.update(components_dtype_dict)
dtype_view[ICON] = 'string'
dtype_view[ID_CHILD] = 'Int32'
dtype_view[ID_PARENT] = 'Int32'
dtype_view[KEYCODE_PARENT] = 'string'
dtype_view[DATE_BEGIN_CE] = 'string'
dtype_view[DATE_END_CE] = 'string'

# Treeview Tabulator tips
dtips_view = {
    KEYCODE: 'A string that identifies the sequence, should be unique.', 
    PROJECT: 'A string that identifies the project associated to the serie. Free text.', 
    SPECIES: 'The species code, use the ITRDB code.', 
    CATEGORY: 'The value is Measure, Mean or Set. Measure is a serie of ring width, Mean is a set of serie with a mean, Set is a set of serie.', 
    SUBCATEGORY: 'The subcategory is a free text. Could be Master Chronology, Regional Chronology, Local Chronology, Site Chronology, etc.', 
    MEANAS_MEASURE: 'The mean is considered as a measure. Detrend is apply on the mean',
    DATE_BEGIN: 'The date of the first ring, in year.', 
    DATE_END: 'The date of the last ring, in year.', 
#    SYNC: 'The date is certain if True else unknown or uncertain.', 
    SITE_LATITUDE: 'The latitude of the site location.', 
    SITE_LONGITUDE: 'The longitue of the site location.', 
    SITE_ELEVATION: 'The elevation of the site location.',
    SITE_CODE: 'The code or name of the site.',
    PITH: 'a boolean that indicates if the pith is present.', 
    #PITH_ABSENT: '?',
    SAPWOOD: 'The position of the sapwood in the serie.', 
    CAMBIUM: 'A boolean that indicates if the cambium is present.', 
    CAMBIUM_SEASON: 'The season (spring or summer) of the cambium.', 
    BARK: 'A boolean that indicates if the bark is present.', 
    CREATION_DATE: 'The date of the creation of the serie.',
    DATE_SAMPLING : 'The date of the sampling of the serie.',
    LABORATORY_CODE : 'The code of the laboratory that has produced the serie.',
    PERS_ID : 'The id of the person that has produced the serie.',
    BIBLIOGRAPHY_CODE : 'A reference to a bibliography.',
    COMMENTS: 'Free text comments.',
    URI: 'A link to a file or a web page that explains the serie.',
    DATA_LENGTH : 'The number of rings in the serie.', 
    DATA_TYPE : 'The type of the data. It must be "RAW" in the tree and a detrend type in other cases.', 
    DATA_VALUES : 'The width of the rings.',
    DATA_WEIGHTS : 'The weight of each ring. 1 for each Tree rings, the number of trees at each Mean rings.', 
    DATA_INFO : 'List of pairs (Id, Offset) for each component in the Mean.',
    DATA_SIGNATURES : 'List of signatures for each component in the Mean.',
    INCONSISTENT : 'A boolean indicating whether the series is inconsistent: data errors/offsets, Mean component changes.',
    TAG: 'Use to group series in a plot.',
    COMPONENT_COUNT: 'Number of components in the Mean.',
}



#--------------------------
# Excel Columns OUTPUT

excel_columns = [KEYCODE, PROJECT, SPECIES, CATEGORY, SUBCATEGORY, MEANAS_MEASURE, DATE_BEGIN, DATE_END,
                 SITE_LATITUDE, SITE_LONGITUDE, SITE_ELEVATION, SITE_CODE, 
                 PITH, SAPWOOD, CAMBIUM, CAMBIUM_SEASON, BARK, 
                 DATA_LENGTH, DATA_TYPE, DATA_SIGNATURES, COMPONENT_COUNT]


#--------------------------
# Statistics

## Statistics fileds
DATA_NAN = 'NbMissingRing'
STAT_MEAN = 'Mean'
STAT_MEDIAN= 'Median'
STAT_MODE= 'Mode'
STAT_STD= 'STD'
STAT_VAR= 'Variance'
STAT_MIN= 'Minimum'
STAT_MAX= 'Maximum'
STAT_PERC25= 'Percentil 25'
STAT_PERC50= 'Percentil 50'
STAT_PERC75= 'Percentil 75'
STAT_SUM= 'Sum'
STAT_KURTOSIS= 'Kurtosis'
STAT_SKEWNESS= 'Skewness'
STAT_ENTROPY= 'Entropy'
#STAT_NORM= 'Norm'

## Statistics type dictonary
stat_dtype_dict = {
    DATA_NAN: 'Int32',
    STAT_MEAN: 'Float32',
    STAT_MEDIAN: 'Float32',
    STAT_MODE: 'Float32',
    STAT_STD: 'Float32',
    STAT_VAR: 'Float32',
    STAT_MIN: 'Float32',
    STAT_MAX: 'Float32',
    STAT_PERC25: 'Float32',
    STAT_PERC50: 'Float32',
    STAT_PERC75: 'Float32',
    STAT_SUM: 'Float32',
    STAT_KURTOSIS: 'Float32',
    STAT_SKEWNESS: 'Float32',
    STAT_ENTROPY: 'Float32',
    #STAT_NORM: 'Float32',
}

detrend_dtype_dict = {
    #PITH_ESTIMATED: 'Int32',
    #PITH_LOWER: 'Int32',
    #PITH_UPPER: 'Int32',
    CAMBIUM_ESTIMATED: 'Int32',
    CAMBIUM_LOWER: 'Int32',
    CAMBIUM_UPPER: 'Int32',
    SLOPE : 'object',
}

#--------------------------
# Package Tabulator
ADD_STATISTICS = False
dtype_package = copy.deepcopy(dtype_view)

dtype_package.update(detrend_dtype_dict)
if ADD_STATISTICS:  
    dtype_package.update(stat_dtype_dict)




""" I/O for Dendron IV format """

__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"
__license__ = "GPL"
import glob
import logging
import copy
import re
from lxml import etree as ET
import numpy as np
import pandas as pd
import pickle
import zipfile
import os
import shutil

from pyDendron.tools.location import reverse_geocode, get_elevation
from pyDendron.dataset import Dataset
from pyDendron.app_logger import logger, perror
from pyDendron.dataname import *
from pathlib import Path

class IODendronIV:
    """
    Represents a DendronIV object.

    Args:
        glob_str (str): The glob string used to search for files.
        source (str): The source of the data.
        laboratories (str): The laboratories associated with the data.
        places (str): The filename of the places file.
        get_place (bool): Flag indicating whether to get place information.
        get_altitude (bool): Flag indicating whether to get altitude information.

    Attributes:
        glob_str (str): The glob string used to search for files.
        uri (str): The URI of the file.
        filename (str): The filename.
        source (str): The source of the data.
        id (int): The index.
        ids (dict): A dictionary containing the indexes.
        sequences (list): A list of sequences.
        components (list): A list of components.
        get_place (bool): Flag indicating whether to get place information.
        get_altitude (bool): Flag indicating whether to get altitude information.
        reverses_location (dict): A dictionary containing reverse locations.
        elevations (dict): A dictionary containing elevations.
        laboratories (str): The laboratories associated with the data.
        places_fn (str): The filename of the places file.
        projects (dict): A dictionary containing the projects.

    Methods:
        _read_places(): Reads the places file.
        _write_places(): Writes the places file.
        _readall(): Reads all the data.
        get_components(seqs, card_dict): Retrieves the components from the given sequences and card dictionary.
        read_begin_table(root): Reads the begin table.
        _add_project(seqs, card_dict): Adds a project.
        _read_cards(root, card_dict): Reads the cards.
        _read_card(card, card_dict): Reads a card.
        _read_field(field, meta): Reads a field.
    """
    file_extension = ['.Dxml']
    
    def __init__(self, source=None, laboratories=None, base_directory='./', places='reverse_places.p', get_place=False, get_altitude=False):
        self.glob_str = '*'+IODendronIV.file_extension[0]
        self.uri = None
        self.filename = None
        self.source = source
        self.id = 0
        self.ids = {}
        self.sequences = []
        self.components = []
        self.get_place = get_place
        self.get_altitude = get_altitude
        self.reverses_location = {}
        self.elevations = {}
        self.laboratories = laboratories
        self.places_fn = places
        self.projects = {}
        self.base_directory = Path(base_directory)
        self.dataset = None

        #self._readall()

    def _read_places(self):
        """
        Reads the places file.
        """
        if (self.places_fn is not None) and Path(self.places_fn).exists():
            with open(self.places_fn , 'rb') as fic:
                self.reverses_location, self.elevations = pickle.load(fic)

    def _write_places(self):
        """
        Writes the places file.
        """
        with open(self.places_fn , 'wb') as fic:
            pickle.dump([self.reverses_location, self.elevations], fic)

    def _readall(self):
        """
        Reads all the data.
        """
        self._read_places()

        #lst = [x for x in glob.glob(self.glob_str, recursive=True)]
        lst = list(self.base_directory.rglob(self.glob_str))
        if len(lst) == 0:
            logger.error(f'No files found with {self.glob_str} in {self.base_directory}')
            return 
        #perror(f"self.base_directory: {self.base_directory}")
        #perror(f"self.glob_str: {self.glob_str}")
        #perror(f"lst: {lst}")
        card_dict = {}
        seqs = {}
        for i, self.filename in enumerate(lst):
            if self.filename.name.startswith('.'):
                continue
            #perror(f"filename: {self.filename}")
            self.uri = Path(self.filename).resolve().as_uri()
            if (i % 25) == 0:
                #logger.debug('_write_places')
                self._write_places()
            with open(self.filename, 'r', encoding='ISO-8859-1') as file:
                    data = file.read()
                    data = data.replace('&', '&amp;')
                    try:
                        root = ET.fromstring(data)
                    except Exception as inst:
                        #perror(f'XML error in {self.filename} {inst} ')
                        logging.warning(f'XML error in {self.filename} {inst} ')
                        root = ET.fromstring(data, ET.XMLParser(recover=True))
                    card_dict.update(self.read_begin_table(root))
                    fn_seqs = self._read_cards(root, card_dict)
                    #for key in fn_seqs:
                    #    perror(f"key: {key} {fn_seqs[key]}")
                    seqs.update(fn_seqs)

        diff =  set(card_dict.values()) - set(seqs.keys())
        if len(diff) > 0:
            logging.warning(f'\t missing sequences  {diff}')

        self.sequences = list(seqs.values())
        self.components = self.get_components(seqs, card_dict)
        self._write_places()

    def get_components(self, seqs, card_dict):
        """
        Retrieves the components from the given sequences and card dictionary.

        Args:
            seqs (dict): A dictionary containing the sequences.
            card_dict (dict): A dictionary containing the card information.

        Returns:
            list: A list of dictionaries representing the components.

        Raises:
            None

        """
        components = []
        for id_parent in seqs:
            meta = seqs[id_parent]
            if CATEGORY not in meta:
                logging.warning(f'\t undefined category for {meta[KEYCODE]}')
            elif (meta[CATEGORY] != MEASURE) and ('comps' in meta):
                tmp = []
                for comp in meta['comps']:
                    if comp in card_dict:
                        id_child = card_dict[comp]
                        if id_child != id_parent:
                            if id_child not in seqs:
                                logging.warning(f'\t remove missing sequences {id_child} in {id_parent}')
                            else:
                                offset = 0
                                if DATE_BEGIN in seqs[id_child]:
                                    offset = seqs[id_child][DATE_BEGIN]
                                    if pd.notna(offset) and (offset > 3000):
                                        ('Offset', offset)
                                        offset -= 3000
                                        seqs[id_child][DATE_BEGIN] = pd.NA
                                        seqs[id_child][DATE_END] = pd.NA

                                tmp.append({ID_PARENT:id_parent, ID_CHILD:id_child, OFFSET:offset})
                if len(tmp) == 0:
                    logging.warning(f"empty set {id_parent} {meta[KEYCODE]} ({meta[CATEGORY]})")
                components += tmp
        return components

    def read_begin_table(self, root):
        """
        Reads the begin table.

        Args:
            root (Element): The root element.

        Returns:
            dict: A dictionary containing the card information.

        Raises:
            None

        """
        card_dict = {}
        data = root.find('FirstTable/text').text
        for line in data.split('\n'):
            line = line.strip()
            fields = line.split()
            if (len(fields) == 4) and (fields[3] == 'Ok'):
                if fields[0] in card_dict:
                    logging.warning(f'\t duplicate card: {fields[0]}')
                card_dict[fields[0]] = self.id
                self.id += 1
        #print('card_dict', card_dict)
        return card_dict

    def _add_project(self, seqs, card_dict):
        """
        Adds a project.

        Args:
            seqs (dict): A dictionary containing the sequences.
            card_dict (dict): A dictionary containing the card information.

        Returns:
            tuple: A tuple containing the project and the components.

        Raises:
            None

        """
        fn = Path(self.filename)
        #project = fn.parent.name
        #perror(f"base_directory: {self.base_directory}")
        #perror(f"fn: {fn}")
        #perror(f"relative path: {fn.relative_to(self.base_directory)}")
        #perror(f"relative path parent: {fn.relative_to(self.base_directory).parent} {fn.relative_to(self.base_directory)}")
        project = fn.relative_to(self.base_directory).parent
        
        if project in self.projects:
            return project, self.projects[project]['comps']
        else:
            file_id = self.id
            meta = {ID: file_id, 
                    KEYCODE: project, 
                    PROJECT: project, 
                    #URI: self.filename,
                    CATEGORY: SET,
                    'comps': {} 
                    }
            seqs[file_id] = meta
            card_dict[file_id] = file_id
            self.id += 1
            self.projects[project] = meta
            return project, meta['comps']

    def _read_cards(self, root, card_dict):
        """
        Reads the cards.

        Args:
            root (Element): The root element.
            card_dict (dict): A dictionary containing the card information.

        Returns:
            dict: A dictionary containing the sequences.

        Raises:
            None

        """
        seqs = {}
        project, comps = self._add_project(seqs, card_dict)

        for card in root.findall("Card"):
            meta, key = self._read_card(card, card_dict)
            #print('meta', meta)
            if meta is not None:
                if meta[ID] in seqs:
                    logging.warning(f'\t duplicate sequences: {key} / {meta[KEYCODE]}')
                else:
                    meta[PROJECT] = project
                    meta[URI] = self.filename
                    seqs[meta[ID]] = meta
                    comps[meta['cardName']] = 'head'
        return seqs

    def _read_card(self, card, card_dict):
        """
        Reads a card.

        Args:
            card (Element): The card element.
            card_dict (dict): A dictionary containing the card information.

        Returns:
            tuple: A tuple containing the metadata and the card name.

        Raises:
            None

        """
        #print("-"*10)
        meta = {}
        card_name = card.find('name').text.strip()
        if card_name == '000_A':
            return None, None
        if card_name not in card_dict:
            logging.warning(f'\t Card not in list {card_name}')
            return None, None
        meta['cardName'] = card_name
        user_name = card.find('userName').text.strip()
        meta[ID] = card_dict[card_name]
        meta[KEYCODE] = user_name
        for field in card.findall("field"):
            self._read_field(field, meta)

        return meta, card_name

    def _read_field(self, field, meta):
        """
        Reads a field.

        Args:
            field (Element): The field element.
            meta (dict): A dictionary containing the metadata.

        Returns:
            None

        Raises:
            None

        """
        def _read_category():
            if text.startswith('indiv'):
                meta[CATEGORY] = MEASURE
            elif text.startswith('group'):
                meta[CATEGORY] = SET
            else:
                logging.warning(f'\t Unknown card type |{text}| {name}')
                meta[CATEGORY] = MEASURE

        def _read_values():
            values = []
            #perror(f"field.text: {text}")
            for line in text.split('\n'):
                line = line.strip()
                if line != '':
                    try:
                        values.append(float(line))
                    except Exception as inst:
                        logging.warning(f'\t* values error: |{line}|')
                else:
                    values.append(np.nan)
            
            #perror(f"values: {values}")
            
            if len(values) > 0:
                meta[DATA_VALUES] = np.array(values)
                meta[DATA_LENGTH] = len(values)
                meta[DATA_TYPE] = RAW
                if (CATEGORY in meta) and (meta[CATEGORY] == SET):
                    meta[CATEGORY] = MEAN

        def _read_location():
            f = text.split()
            try:
                meta[SITE_LATITUDE] = float(f[1])
                meta[SITE_LONGITUDE] = float(f[0])
            except Exception as inst:
                meta[SITE_LATITUDE] = pd.NA
                meta[SITE_LONGITUDE] = pd.NA
            meta[SITE_ELEVATION] = pd.NA
            meta[SITE_CODE] = ''
            if self.get_place:
                _, __, ___, ____, _____, meta[SITE_CODE] = reverse_geocode(meta[SITE_LATITUDE], meta[SITE_LONGITUDE], self.reverses_location)
            if self.get_altitude: 
                meta[SITE_ELEVATION] = get_elevation(meta[SITE_LATITUDE], meta[SITE_LONGITUDE], self.elevations)

        def _read_working_list():
            #f = text.split()
            comps = {}
            for line in text.split('\n'):
                line = line.strip()
                f = line.split()
                if len(f) == 5:
                    comps[f[1]] = f
            meta['comps'] = comps

        name = field.find('name').text.strip()
        text = field.find('text').text.strip()

        if name == 'card type':
            _read_category()
        elif name == 'i_01':
            _read_values()
        elif name =='sample':
            v = text.split('\n')
            if len(v) > 3:
                if v[0] == '1':
                    meta[PITH] = True
                i = re.search(r'\d+', v[1])
                if i:
                    meta[SAPWOOD] = int(i.group()) - 1
                if v[2] == '1':
                    meta[CAMBIUM] = True
        elif (name == 'material') and (text != ''):
            meta[SUBCATEGORY] = text
        elif (name == 'creation date') and (text != ''):
            try:
                meta[CREATION_DATE] = pd.to_datetime(text, format="%Y%m%d")
            except Exception as inst:
                meta[CREATION_DATE] = ''
        elif (name == 'start date') and (text != ''):
            meta[DATE_BEGIN] = float(text)
        elif (name == 'stop date') and (text != ''):
            meta[DATE_END] = float(text)
        elif (name == 'location') and (text != ''):
            _read_location()
        elif (name == 'tree type') and (text != ''):
            meta[SPECIES] = text
        elif name == 'working list':
            _read_working_list()
        elif name == 'waiting list':
            if text != '':
                meta[DATA_INFO] = text
        elif name == 'comments':
            if text != '':
                meta[COMMENTS] = text
        elif name == 'labs authors':
            if text != '':
                meta[LABORATORY_CODE] = text
        elif name == 'people authors':
            if text != '':
                meta[PERS_ID] = text
        elif name == 'references':
            if text != '':
                meta[BIBLIOGRAPHY_CODE] = text
        elif name == 'source file':
            if text != '':
                meta[URI] = text
                    
    def to_dataset(self, keycode_parent='Dataset', root_id=ROOT, trash_keycode='Trash', workshop_keycode='Workshop', clipboard_keycode='Clipboard'):
        """
        Converts the current state of the object to a Dataset object.

        Args:
            keycode_parent (str): The keycode for the root node of the dataset. Default is 'Dataset'.
            root_id (int): The index of the root node. Default is ROOT.
            trash_keycode (str): The keycode for the trash node of the dataset. Default is 'Trash'.
            workshop_keycode (str): The keycode for the workshop node of the dataset. Default is 'Workshop'.
            clipboard_keycode (str): The keycode for the clipboard node of the dataset. Default is 'Clipboard'.

        Returns:
            Dataset: The converted Dataset object.
        """
        if len(self.sequences) == 0:
            return None
        
        self.dataset = Dataset(sequences=self.sequences, components=self.components, save_auto=False)

        for k, t in sequences_dtype_dict.items():
            if k not in self.dataset.sequences.columns:
                self.dataset.sequences[k] = ''
            if t == 'string':
                self.dataset.sequences[k] = self.dataset.sequences[k].fillna('')

        self.dataset.sequences[INCONSISTENT] = False
        self.dataset.sequences[SPECIES] = self.dataset.sequences[SPECIES].str.upper()
        mask = self.dataset.sequences[SPECIES].str.startswith('QU')
        self.dataset.sequences.loc[mask, SPECIES] = 'QU'

        logger.info(f'to_dataset roots: {self.dataset.get_roots()}')
        self.dataset.new_root(keycode_parent, root_id)
        self.dataset.new_clipboard(clipboard_keycode)
        self.dataset.new_trash(trash_keycode)
        self.dataset.new_workshop(workshop_keycode)

        return self.dataset
    
    def compact(self, drop_duplicate=False, mean_to_set=False):
        """
        Remove duplicate samples in the tree and keep deeper nodes.
        Parent categories are set to SET and RING_* are set empty.

        Parameters:
        - drop (bool): If True, remove the duplicate samples from the dataset. Default is False.

        Returns:
        - dataset: The modified dataset after removing duplicate samples.
        """
        def iterate(node):
            dup = [x for (x, y) in node.detect_duplicates(category=MEASURE, raise_error=False)]
            for child in node.children:
                if child.category != MEASURE:
                    iterate(child)
                else:
                    if child.id in dup:
                        remove_list.append((node.id, child.id))
                        mean_id_list.append(node.id)

        if self.dataset is None:
            return None
        
        if drop_duplicate:
            tree = self.dataset.get_descendants(self.dataset.get_roots())
            remove_list = []
            mean_id_list = []
            iterate(tree)
            #logger.info(f'compact remove_list: {remove_list}')
            self.dataset.components.drop(remove_list, inplace=True)
            
        if mean_to_set:
            mask = self.dataset.sequences.index.isin(set(mean_id_list))
            cols = [CATEGORY, DATA_LENGTH, DATA_INFO, DATA_TYPE, DATA_WEIGHTS, DATA_VALUES]
            self.dataset.sequences.loc[mask, cols] = SET, pd.NA, pd.NA, pd.NA, pd.NA, pd.NA

        return self.dataset
    
    def project_to_set(self, id_root=ROOT):
        def create_set(path, project_sets):
            id_parent = [id_root]
            #perror(f"path: {path}")
            for p in path.parts:
                #perror(f" p: {p}")
                if p not in project_sets:
                    self.id += 1
                    #perror(f"   new p: {p} {id_parent[-1]} {self.id}")
                    self.dataset.new(keycode=p, category=SET, id_parent=id_parent[-1], id=self.id)
                    id_parent.append(self.id)
                    project_sets[p] = copy.copy(id_parent)
                else:
                    id_parent = copy.copy(project_sets[p])
                    #perror(f"   get p: {p} {id_parent[-1]}")
            return id_parent
        
        if self.dataset is None:
            return None

        project_sets = {}
        for project, df in self.dataset.sequences.groupby(PROJECT):
            path = Path(project)
            last_project = create_set(path, project_sets)
            child_mask = self.dataset.components.index.get_level_values(ID_CHILD).isin(df.index)
            root_mask = self.dataset.components.index.get_level_values(ID_PARENT).isin([id_root])
            comps = self.dataset.components[child_mask & root_mask]
            #perror(f"last_project: {path} {last_project} {len(comps)}")
            #perror(f"comps.to_records: {comps.to_records(index=True)}")
            self.dataset.move(comps.to_records(index=True), last_project)
            
            
    
    def read(self, directory, keycode_parent='Dendron-IV'):
        self.base_directory = Path(directory)
        if not  self.base_directory.is_dir():
            raise ValueError(f'{ self.base_directory} is not a directory')
        #self.glob_str = '*.Dxml'
        
        #root_keycode = self.base_directory.parts[-1]
        self._readall()
        self.to_dataset(keycode_parent=keycode_parent)
        
        self.compact(drop_duplicate=True, mean_to_set=True)
        self.project_to_set()
        
        return self.dataset

    def read_buffer(self, keycode_parent, buffer, mine_type=None):
        if mine_type != 'application/zip':
            raise ValueError(f'Unsupported mine_type {mine_type}')
        
        output_directory = Path(self.base_directory) / keycode_parent
        os.makedirs(output_directory, exist_ok=True)
        
        output_file = output_directory / Path(keycode_parent).with_suffix(".zip")
        with open(output_file, 'wb') as f:
            f.write(buffer)  

        zip = zipfile.ZipFile(output_file)
        zip.extractall(output_directory)
        
        dataset = self.read(output_directory, keycode_parent=keycode_parent) 
        shutil.rmtree(output_directory)
        return dataset             
